import torch
import numpy as np
import itertools
from scipy import stats
import copy
import seaborn as sns
import matplotlib.pyplot as plt
from matplotlib import colors
from NaroNet.BioInsights.add_annotation_stat import add_stat_annotation
import pandas as pd
from itertools import combinations
from scipy.cluster.hierarchy import fcluster
from sklearn.cluster import KMeans
import random
from scipy.cluster.hierarchy import ward, fcluster
from scipy.spatial.distance import pdist
from scipy.stats import mannwhitneyu
from NaroNet.utils.parallel_process import parallel_process
from NaroNet.BioInsights.adjust_pvalues import lsu
from NaroNet.BioInsights.adjust_pvalues import hochberg
from NaroNet.BioInsights.adjust_pvalues import holm_bonferroni
from NaroNet.BioInsights.adjust_pvalues import sidak

def differential_abundance_analysis(dataset, heatmapALL, labels, IndexAndClass,isTraining):        
        pvalueThreshold=1.25#pvalueThreshold=0.0000000001
        statisticalTest = [] # List of columns with statistical difference. 1.p-value, 2.Cluster step, 3.column of the heatmap.
        
        class Model(torch.nn.Module):
            def __init__(self, num_features, n_classes):
                super(Model, self).__init__()
                # self.lin1 = torch.nn.Linear(num_features, 1024, bias=True)                
                self.lin4 = torch.nn.Linear(num_features, n_classes, bias=True)

            def reset_parameters(self):
                # self.lin1.reset_parameters()                
                self.lin4.reset_parameters()

            def forward(self,features, labels):
                # UM = torch.nn.functional.relu(self.lin1(features)) # torch.cat(self.X,dim=-1)            
                return self.lin4(features)         

        class Load_Model(torch.nn.Module):
            def __init__(self,dataset):
                super(Load_Model, self).__init__()      
                model = torch.load(dataset.processed_dir_cross_validation+'model.pt')                          
                self.lin1 = model[0] 
                self.batchNorm = model[1] 
                self.lin2 = model[2]

            def reset_parameters(self):
                return
                # self.lin1.reset_parameters()                
                # self.lin4.reset_parameters()

            def forward(self,features, labels):
                UM = torch.nn.functional.relu(self.lin1(features)) # torch.cat(self.X,dim=-1)   
                UM = self.batchNorm(UM)       
                return self.lin2(UM)         
        
        class Model_NonLin(torch.nn.Module):
            def __init__(self, num_features, n_classes):
                super(Model_NonLin, self).__init__()
                self.lin1 = torch.nn.Linear(num_features, 1024, bias=True)                
                self.lin4 = torch.nn.Linear(1024, n_classes, bias=True)

            def reset_parameters(self):
                # self.lin1.reset_parameters()                
                self.lin4.reset_parameters()

            def forward(self,features, labels):
                UM = torch.nn.functional.relu(self.lin1(features)) # torch.cat(self.X,dim=-1)            
                return self.lin4(UM)      
            
            # loss = F.cross_entropy(UM, labels[:,0].long())     
            # pred = UM.max(1)[1].detach().cpu().numpy()

        def trainModel(model,dependentVariable,labels,model_is_loaded):
            # Initialize Model
            if not model_is_loaded:
                model.reset_parameters() 
                n_iter = 10000    
            else:
                n_iter = 5
            optimizer = torch.optim.Adam(model.parameters(), lr=0.0001,weight_decay=0.0001)
            dependentVariable = torch.tensor(dependentVariable,device='cpu',dtype=torch.float32)
            labels = torch.tensor(labels,device='cpu',dtype=torch.float32)
            for i in range(n_iter):
                losses = model(dependentVariable,labels) 
                # pred = losses.max(1)[1].detach().cpu().numpy()                               
                # print('train_acc:',np.equal(pred,labels.cpu().numpy()).mean())
                # l = torch.nn.functional.cross_entropy(losses.squeeze(), labels)
                l = torch.nn.functional.cross_entropy(losses, labels.long())     
                l.backward()
                optimizer.step()
                # print(l.cpu().detach().numpy())
            pred = losses.max(1)[1].detach().cpu().numpy()            
            return model.eval(), np.equal(pred,labels.cpu().numpy()).mean(), np.stack([torch.nn.functional.cross_entropy(losses[i,:].unsqueeze(dim=0), labels[i].unsqueeze(dim=0).long()).detach().cpu().numpy() for i in range(labels.shape[0])])
        
        def validateModel(model, dependentVariable, labels):
            dependentVariable = torch.tensor(dependentVariable,device='cpu',dtype=torch.float32)
            labels = torch.tensor(labels,device='cpu',dtype=torch.float32)
            losses = model(dependentVariable,labels)  
            labels_c = labels.cpu().numpy()
            
            # Obtain the prediction confidence of the predicted class
            r_loss_pred_class = [i.max() for n,i in enumerate(torch.nn.functional.softmax(losses,dim=1).detach().cpu().numpy())]           
            r_loss_real_class = [i[int(labels_c[n])] for n,i in enumerate(torch.nn.functional.softmax(losses,dim=1).detach().cpu().numpy())]
            restricted_loss = [i[int(labels_c[n])] for n,i in enumerate(losses.detach().cpu().numpy())]
            return np.equal(losses.max(1)[1].detach().cpu().numpy(),labels_c).mean(), np.array(restricted_loss), np.array(r_loss_pred_class), np.array(r_loss_real_class), losses.max(1)[1].detach().cpu().numpy()
            
        def findMinimumNumberOfClusters(dependentVariable,model,labels,Unrestricted_loss,pvalueThreshold):
            # Indices to select unrestricted model.
            indices = np.array(list(range(dependentVariable.shape[1])))
            # Possible Combinations
            possible_comb = list(range(dependentVariable.shape[1]))

            # Find the minimum number of significant clusters.
            for n_significant_clusters in range(dependentVariable.shape[1]): 
                
                # Save p-value of statistical tests.
                if n_significant_clusters==0:
                    stats_indices = []
                elif n_significant_clusters==1:
                    possible_comb = sorted(range(len(stats_indices)), key=lambda i: stats_indices[i])[-7:]
                    
                # Iterate n_significant_clusters, and all their possible combinations.
                for selectedCluster in itertools.combinations(possible_comb,n_significant_clusters+1):                                
                    # Execute Restricted model
                    Restricted_dependentVariable = np.zeros(dependentVariable.shape)
                    for c in selectedCluster:
                        Restricted_dependentVariable[:,indices==c] = dependentVariable[:,indices==c]
                    ACC, Restricted_loss, pred_Confidence, real_class_confidence, pred_Label = validateModel(model,Restricted_dependentVariable,labels)
                    print('Restricted Model accuracy: ',ACC)

                    # Check if unrestricted model is significantly different to restricted model.
                    if stats.shapiro(Restricted_loss)[1]<0.05 or stats.shapiro(Unrestricted_loss)[1]<0.05:
                        # Normality is rejected
                        if len(selectedCluster)==1:
                            stats_indices.append(stats.kruskal(Restricted_loss,Unrestricted_loss)[1])
                        if stats.kruskal(Restricted_loss,Unrestricted_loss)[1]>pvalueThreshold:
                            return selectedCluster, stats.kruskal(Restricted_loss,Unrestricted_loss)[1] # This unrestricted mdel is not sinificantly different to the restricted one.
                    else: 
                        # Normality is accepted
                        if len(selectedCluster)==1:
                            stats_indices.append(stats.ttest_rel(Restricted_loss,Unrestricted_loss)[1])
                        if stats.ttest_rel(Restricted_loss,Unrestricted_loss)[1]>pvalueThreshold:
                            return selectedCluster, stats.ttest_rel(Restricted_loss,Unrestricted_loss)[1]  # This unrestricted mdel is not sinificantly different to the restricted one.

        def is_bad_combination(possible_ind, selectedCluster):
            for ind in possible_ind:
                num_pos=0
                for i in ind[1]:
                    num_pos += sum([s==i for s in selectedCluster])
                if num_pos==len(ind[1]):                            
                    return True
            return False
        
        def findSignificantClusters(dataset,dependentVariable,model,labels,Unrestricted_loss,pred_confidence_Unrestricted,pvalueThreshold):
            # Indices to select unrestricted model.
            indices = np.array(list(range(dependentVariable.shape[1])))
            # Possible Combinations
            possible_comb = list(range(dependentVariable.shape[1]))
            possible_ind = []
            Indices = list(range(dependentVariable.shape[1]))

            # Save Statistics
            stats_indices = []  
            
            # Find the minimum number of significant clusters.
            for n_significant_clusters in range(2): #dependentVariable.shape[1]                                                                 
              
                # Iterate n_significant_clusters, and all their possible combinations.
                for selectedCluster in itertools.combinations(Indices,n_significant_clusters+1):                                
                    
                    # Check if model should be checked
                    if not is_bad_combination(possible_ind, selectedCluster) and n_significant_clusters>0:
                        continue
                    
                    # Execute Restricted model
                    Restricted_dependentVariable = copy.deepcopy(dependentVariable)
                    for c in selectedCluster:
                        Restricted_dependentVariable[:,indices==c] = 0
                    ACC, Restricted_loss, pred_Confidence_restricted, real_class_confidence, pred_Label = validateModel(model,Restricted_dependentVariable,labels)
                    # print('Restricted Model accuracy: ',ACC)

                    # Check if unrestricted model is significantly different to restricted model.
                    # if stats.shapiro(Restricted_loss)[1]<0.05 or stats.shapiro(Unrestricted_loss)[1]<0.05:
                        # Normality is rejected
                        # if len(selectedCluster)==1:
                            # stats_indices.append(stats.kruskal(Restricted_loss,Unrestricted_loss)[1])                    
                    if np.mean(pred_confidence_Unrestricted)>np.mean(pred_Confidence_restricted):               
                        stats_indices.append([stats.kruskal(Restricted_loss,Unrestricted_loss)[1],selectedCluster,ACC,[i if i<=3 else 3 for i in pred_confidence_Unrestricted/pred_Confidence_restricted]])
                    else:
                        stats_indices.append([1,selectedCluster,ACC,[i if i<=3 else 3 for i in pred_confidence_Unrestricted/pred_Confidence_restricted]])
                    
                    # else: 
                        # Normality is accepted
                        # if len(selectedCluster)==1:
                            # # stats_indices.append(stats.ttest_rel(Restricted_loss,Unrestricted_loss)[1])
                        # if stats.ttest_rel(Restricted_loss,Unrestricted_loss)[1]<pvalueThreshold:
                            # return selectedCluster, stats.ttest_rel(Restricted_loss,Unrestricted_loss)[1]  # This unrestricted mdel is not sinificantly different to the restricted one.
                
                # Order Indices to show wich are the most important
                if n_significant_clusters==0:
                    Indices = sorted(range(len(stats_indices)), key=lambda i: stats_indices[i][0])[:int(len(stats_indices)*1)]    
                
                # Save p-value of statistical tests.                                
                possible_comb = sorted(range(len(stats_indices)), key=lambda i: stats_indices[i][0])[:int(len(stats_indices)*1)]
                possible_ind = [stats_indices[i] for i in possible_comb]            
                
                # Do the statistical test
                # print('BioInsights: 'stats_indices[possible_comb[0]][0])
                if stats_indices[possible_comb[0]][0]<pvalueThreshold:
                    break

            # Find the top 1 significant cluster per patient
            Top1PIRPerPatient=[]
            for i in range(Restricted_loss.shape[0]):
                Top1PIRPerPatient.append([(),0])
            for sts in stats_indices:
                if sts[0]<pvalueThreshold or (not 'Synthetic' in dataset.root):
                    for n, pat in enumerate(Top1PIRPerPatient):
                        if pat[1]<sts[3][n]:
                            pat[1]=sts[3][n]
                            pat[0]=sts[1]

            return stats_indices[possible_comb[0]][1], stats_indices[possible_comb[0]][0], stats_indices[possible_comb[0]][2],[stats_indices[i] for i in possible_comb], Top1PIRPerPatient# This unrestricted mdel is not sinificantly different to the restricted one.        

        def obtain_relevant_PIR(Top1PerPatient,topkClustersStats):
            '''
                Returns PIR statistics per patient
            '''

            # Obtain unique cell types that are significnat
            Top1PerPatientUniq = list(set([i[0] for i in Top1PerPatient]))
            top1ClustersStats=[]
            for topk in topkClustersStats:
                if any([len(set(uniq) & set(topk[1]))==len(uniq) for uniq in Top1PerPatientUniq]) or topk[0]<0.05:
                    top1ClustersStats.append(topk)

            # Create Prediction Index in Restricted Model for the statistically significant cell type combinations
            stats_sign = []
            RestrictedModelData={}
            for topk in top1ClustersStats:
                Significant_Clust_str = ''
                for S_C_S in topk[4]:
                    Significant_Clust_str += '+'+S_C_S[1]
                if 0.01<topk[0]<0.05:
                    RestrictedModelData[Significant_Clust_str[1:]+'*'] = topk[3]
                elif 0.001<topk[0]<0.01:
                    RestrictedModelData[Significant_Clust_str[1:]+'**'] = topk[3]
                elif topk[0]<0.001:
                    RestrictedModelData[Significant_Clust_str[1:]+'***'] = topk[3]
                else:
                    RestrictedModelData[Significant_Clust_str[1:]] = topk[3]

            return top1ClustersStats, RestrictedModelData

        def save_dict_to_excel(dictionary, filename):
            df = pd.DataFrame(dict([(k,pd.Series(v)) for k,v in dictionary.items()])) 
            df.to_excel(filename)

        def showPatientPredictionConfidence(dataset,pred_Confidence, labels,Subgroups_of_patients_labels_names, Subgroups_of_patients_labels,topkClustersStats,IndexAndClass,Top1PerPatient,real_class_confidence):

            list_labels = [l[0] for l in labels[list(labels.keys())[0]]]
                        
            # Visualize prediction confidence and PIR
            pred_Confidence_dict = {}
            pred_Confidence_dict['Subject prediction confidence'] = real_class_confidence      
            pred_Confidence_dict['Subject Name'] = [i[0] for i in IndexAndClass]
            pred_Confidence_dict['Subject Index'] = [i[1] for i in IndexAndClass]
            pred_Confidence_dict['Class'] = [i[2][0] for i in IndexAndClass]
            
            save_dict_to_excel(pred_Confidence_dict, dataset.bioInsights_dir_abundance_analysis+'Prediction_Confidence.xlsx')                        
            # df_PC =  pd.DataFrame.from_dict(pred_Confidence_dict)            
            # df_PIR = pd.DataFrame.from_dict(RestrictedModelData)

            # Order Subgroups_of_patients_labels
            # listt=list(range(max(set(Subgroups_of_patients_labels))+1))
            # strt = 0
            # for i in set(Subgroups_of_patients_labels):
            #     listt[i] = strt
            #     strt+=1
            # Subgroups_of_patients_labels = [listt[spl] for spl in Subgroups_of_patients_labels]
            # cp = sns.color_palette("bright",len(set(Subgroups_of_patients_labels)))
            # Subgroups_of_patients_labels_colors = [cp[spl-1] for spl in Subgroups_of_patients_labels]
            # names = list(Subgroups_of_patients_labels_names.keys()) 
            
            # # Display PIR's per patient.
            # plt.close('all')
            # plt.figure()
            # labels_pal = sns.cubehelix_palette(len(set(list_labels)), light=.9, dark=.1, reverse=True, start=0, rot=-2)
            # labels_lut = dict(zip(map(str, list(set(list_labels))), labels_pal))        
            # labels_colors = [labels_lut[str(k)] for k in list_labels]
            # cm_PIR = sns.clustermap(df_PIR,col_cluster=False,row_cluster=True, xticklabels=1, vmin=1, vmax=1.3, row_colors=[Subgroups_of_patients_labels_colors,labels_colors], linewidths=0, cmap="magma")

            # plt.close('all')
            # plt.figure()
            # labels_pal = sns.cubehelix_palette(len(set(list_labels)), light=.9, dark=.1, reverse=True, start=0, rot=-2)
            # labels_lut = dict(zip(map(str, list(set(list_labels))), labels_pal))        
            # labels_colors = [labels_lut[str(k)] for k in list_labels]
            # cm_PIR = sns.clustermap(df_PIR,col_cluster=False,row_cluster=True, yticklabels=True, vmin=1, vmax=1.4,colors_ratio=0.01,dendrogram_ratio=0.1, row_colors=[labels_colors,Subgroups_of_patients_labels_colors], linewidths=0, cmap="magma",cbar_pos=(int(len(IndexAndClass)/20), .2, .03, .4))
            # hm = cm_PIR.ax_heatmap.get_position()            
            # plt.setp(cm_PIR.ax_heatmap.yaxis.get_majorticklabels(), fontsize=2)
            # cm_PIR.ax_heatmap.set_position([hm.x0, hm.y0, hm.width*0.1, hm.height])
            # for label in range(len(set(Subgroups_of_patients_labels))):
            #     cm_PIR.ax_col_dendrogram.bar(0, 0, color=cp[label], label=names[label], linewidth=0)            
            # cm_PIR.ax_col_dendrogram.legend(title=dataset.experiment_label[0], ncol=4,bbox_to_anchor=(0, 1), loc='upper right')
            # # col = cm_PIR.ax_col_dendrogram.get_position()
            # # cm_PIR.ax_col_dendrogram.set_position([col.x0, col.y0, col.width*0.25, col.height*0.5])            
            # # cm_PIR.ax_heatmap.set_xticklabels(cm_PIR.ax_heatmap.get_xticklabels(), rotation=80,fontsize=3)    
            # plt.savefig(dataset.bioInsights_dir_abundance_analysis_global+'clustermap_Predictive_influence_ratio_per_patient_iter_{}.png'.format(str(dataset.args['epochs'])),dpi=1000)            

            # # Display Prediction Index in Unrestricted Model
            # plt.close('all')
            # plt.figure()
            # real_class_confidence_clrs = [(0,1,0) if rcc>1/len(set(list_labels)) else (1,0,0) for rcc in real_class_confidence]
            # cm = sns.clustermap(pred_Confidence,col_cluster=False,row_cluster=True,vmin=0,vmax=1, 
            #                     row_colors=[labels_colors,real_class_confidence_clrs], linewidths=0, cmap="RdYlGn",
            #                     row_linkage=cm_PIR.dendrogram_row.linkage)#,figsize=(10,20))                                                
            # hm = cm.ax_heatmap.get_position()            
            # plt.setp(cm.ax_heatmap.yaxis.get_majorticklabels(), fontsize=2)
            # cm.ax_heatmap.set_position([hm.x0, hm.y0, hm.width*0.25, hm.height])
            # for label in list(set(list_labels)):
            #     cm.ax_col_dendrogram.bar(0, 0, color=labels_lut[label], label=label, linewidth=0)
            # cm.ax_col_dendrogram.legend(title=dataset.experiment_label[0], loc="best", ncol=5, bbox_to_anchor=(0.47, 0.8), bbox_transform=plt.gcf().transFigure)
            #     col = cm.ax_col_dendrogram.get_position()
            #     cm.ax_col_dendrogram.set_position([col.x0, col.y0, col.width*0.25, col.height*0.5])            
            # cm.savefig(dataset.bioInsights_dir_abundance_analysis+'heatmap_PredictionIndex_iter_{}.png'.format(str(dataset.args['epochs'])),dpi=1000) 

            return  

        def visualizeRegressionAnalysis_(dataset, dependentVariable,clust,list_labels):
            '''
                Compare patient groups using violinplot with umann whitney with bonferrini correction
            '''
            for nn, clust_n in enumerate(clust[1]):                
                # Extract cell type abundance for each patient type
                data = {}
                for lbl in list(set(list_labels)):                    
                    data[lbl] = dependentVariable[:,clust_n][[True if l==lbl else False for l in list_labels]]

                # Check if values are equal                
                for v in data.keys():
                    if len(set(data[v])) <= 1:                    
                        data[v] = np.random.uniform(0,0.00000001,data[v].shape)
                    # data[v][data[v]>2] = (data[v][data[v]>2]/10) + 2
                    # data[v][data[v]<2] = (data[v][data[v]<2]/10) + 2
                
                # Display and save boxplots of abundance.
                plt.close('all')
                plt.figure()                                
                df = pd.DataFrame(dict([ (k,pd.Series(v)) for k,v in data.items() ]))
                labels_pal = sns.cubehelix_palette(len(set(list_labels)), light=.9, dark=.1, reverse=True, start=0, rot=-2)                
                ax = sns.boxplot(data=df, order=list(set(list_labels)),palette=labels_pal, linewidth=2.5)
                ax = sns.stripplot(data=df, order=list(set(list_labels)),color=".5")
                ax.tick_params(labelsize=10)
                ax.set_xticklabels(list(set(list_labels)), fontsize=12)
                ax.set_ylabel("Relative abundance of "+clust[4][nn][1], fontsize=18)                
                test_results_, test_results = add_stat_annotation(ax, data=df, x=None, y=None, order=list(set(list_labels)),
                                                box_pairs=list(combinations(list(set(list_labels)),2)),
                                                test='Mann-Whitney', text_format='star',
                                                loc='inside', verbose=0)
                ax.set_title('(p-value='+str(test_results[0].pval)+')', fontsize=18)
                plt.savefig(dataset.bioInsights_dir_abundance_analysis_global+'/Global_ViolinPlot(Pairs)_'+clust[4][nn][1]+'_iter_'+str(dataset.args['epochs']),dpi=600)                    

            # Extract cell type abundance for each patient type
            data = {'Patch characterization':[],'Relative abundance':[],dataset.experiment_label[0]:[]}       
            for lbl in list(set(list_labels)):     
                for n_c in range(len(clust[1])):  
                    for d in dependentVariable[:,clust[1][n_c]][[True if l==lbl else False for l in list_labels]]:
                        data['Patch characterization'].append(clust[4][n_c][1])            
                        data['Relative abundance'].append(d)
                        data[dataset.experiment_label[0]].append(lbl)                                
            boxpairs=[]
            for n_c in [i[1] for i in clust[4]]:  # For each microenvironment element        
                for lb in combinations(list(set(list_labels)),2): 
                    boxpairs.append(tuple([(n_c,labl) for labl in lb]))
            
            # Check if values are equal                               
            if len(set(data['Relative abundance'])) <= 1:
                data['Relative abundance'] = np.random.uniform(0,0.00000001,len(data['Relative abundance']))

            # Display and save boxplots of abundance.                
            plt.close('all')
            plt.figure()
            df = pd.DataFrame(dict([ (k,pd.Series(v)) for k,v in data.items() ]))
            labels_pal = sns.cubehelix_palette(len(set(list_labels)), light=.9, dark=.1, reverse=True, start=0, rot=-2)
            ax = sns.violinplot(x='Patch characterization', y='Relative abundance', hue=dataset.experiment_label[0],data=df,palette=labels_pal)                
            test_results_, test_results = add_stat_annotation(ax, data=df,x='Patch characterization', y='Relative abundance',hue=dataset.experiment_label[0],
                                            box_pairs=boxpairs,
                                            test='Mann-Whitney', text_format='star',
                                            loc='inside', verbose=0)  
            ax.set_title('(p-value='+str(test_results[0].pval)+')', fontsize=18)
            plt.savefig(dataset.bioInsights_dir_abundance_analysis_global+'/Global_ViolinPlot_'+'_'.join([i[1] for i in clust[4]])+'_iter_'+str(dataset.args['epochs']),dpi=600)
            return 'Done'

        def visualizeRegressionAnalysis(dataset,dependentVariable,labels,pred_Label,topkClustersStats,IndexAndClass,top1):                                    

            dependentVariable = np.concatenate(tuple([dependentVariable[h] for h in dependentVariable]),1)
            
            list_labels = [l[0] for l in labels[list(labels.keys())[0]]]
            list_labels_int = [list(set(list_labels)).index(l) for l in list_labels]
                
            # Prepare parallel process
            dict_subjects = []
            for n, clust in enumerate(topkClustersStats):                
                dict_subjects.append({'dataset':dataset,'dependentVariable':dependentVariable,'clust':clust,'list_labels':list_labels})
            # dict_subjects= dict_subjects[:20]
            # select_patches_from_cohort
            result = parallel_process(dict_subjects,visualizeRegressionAnalysis_,use_kwargs=True,front_num=len(dict_subjects),desc='BioInsights: Save abundance differences between TMEs') 
            
        def showGroups_Patients(RestrictedModelData,labels,dependentVariable,Top1PerPatient,top1ClustersStats,dataset,IndexAndClass):
            '''
                docstring
            '''

            # Calculate dataframe from model data
            PIR_df = pd.DataFrame.from_dict(RestrictedModelData)

            # Perform groups of patients per class.
            list_labels = [l[0] for l in labels[list(labels.keys())[0]]]
            
            # Patient Sub-grouping
            patientSubgroups = {dataset.experiment_label[0]:[],'Patient subgroup':[],'Microenvironment element':[],'PIR value':[],'Patient_index':[],'Patient_name':[],'Centroid_Values: '+'_'.join(PIR_df.keys()):[]}
            Subgroups_of_patients_labels = [0 for i in IndexAndClass]
            for lbl_n, lbl in enumerate(list(set(list_labels))):
                # Extract list of patients for a certain class
                lbl_P = PIR_df[[True if lbl==l else False for l in list_labels]]
                patient_list = [l for l in range(len(list_labels)) if lbl==list_labels[l]]
                if np.array(lbl_P).shape[0]>1:
                    # Perform clustering on patients from a certain class             
                    labels_ = fcluster(ward(pdist(np.array(lbl_P))), t=max(int(lbl_P.shape[0]/18),1), criterion='maxclust')-1            
                    labels_cluster_center = [np.array(lbl_P)[labels_==i].mean(0) for i in range(len(set(labels_)))]
                    # kmeans = KMeans(n_clusters=max(int(lbl_P.shape[0]/18),1), random_state=0).fit(lbl_P)
                    # Assign classes to original list of patients
                    for p_n, p in enumerate(patient_list):
                        Subgroups_of_patients_labels[p] = (labels_[p_n]+(100**lbl_n))
                    # Gather subgroup information in dataframe
                    for subgr_lbl in range(len(set(labels_))):
                        for micro_el in lbl_P.keys():                        
                            for indx, pir_value in enumerate(lbl_P[micro_el][labels_==subgr_lbl]):                                                  
                                patientSubgroups[dataset.experiment_label[0]].append(lbl)
                                patientSubgroups['Patient subgroup'].append(subgr_lbl)
                                patientSubgroups['Microenvironment element'].append(micro_el)
                                patientSubgroups['PIR value'].append(pir_value)   
                                patientSubgroups['Patient_index'].append(lbl_P[micro_el][labels_==subgr_lbl].axes[0][indx]),
                                patientSubgroups['Patient_name'].append(IndexAndClass[lbl_P[micro_el][labels_==subgr_lbl].axes[0][indx]][0])  
                                patientSubgroups['Centroid_Values: '+'_'.join(lbl_P.keys())].append(labels_cluster_center[subgr_lbl])

            # Show heatmap with differences between kmeans centroid
            TME_Names = [k.split('*')[0] for k in PIR_df.keys()]
            TME_Names_Raw = [k for k in PIR_df.keys()]
            listOfIndicesForTME = [t[1] for t in top1ClustersStats]
            TME_Names_oneByone = []
            heatmap_TME = {}
            for TME in TME_Names:
                for T in TME.split('+'):
                    TME_Names_oneByone.append(T)
                    heatmap_TME[T] = []
            heatmap_centroid = {} 
            for PG in set(patientSubgroups[dataset.experiment_label[0]]):
                locate_PG = np.array(patientSubgroups[dataset.experiment_label[0]])==PG
                SubGroups = set(np.array(patientSubgroups['Patient subgroup'])[locate_PG])
                for SG in SubGroups:
                    locate_PG_and_PS = np.logical_and(np.array(patientSubgroups['Patient subgroup'])==SG,locate_PG)
                    Centroid_Values = np.array(patientSubgroups['Centroid_Values: '+'_'.join(PIR_df.keys())])[locate_PG_and_PS][0]
                    if len(Centroid_Values)==1:
                        heatmap_centroid[dataset.experiment_label[0]+':'+PG+'-PS'+str(SG+1)]=list(Centroid_Values)                                        
                    else:
                        heatmap_centroid[dataset.experiment_label[0]+':'+PG+'-PS'+str(SG+1)]=list(stats.zscore(Centroid_Values,nan_policy='omit'))                                        
                    indTME = 0
                    for n_TME, TME in enumerate(TME_Names):
                        for n_TME_un, TME_un in enumerate(TME.split('+')):
                            Patient_indexes = np.unique(np.array(patientSubgroups['Patient_index'])[locate_PG_and_PS])
                            heatmap_TME[TME_Names_oneByone[indTME]].append(np.mean(dependentVariable[Patient_indexes,listOfIndicesForTME[n_TME][n_TME_un]]))
                            indTME+=1

            # TOP-10 PIRs for each TME.
            topk = 10
            ALLPatient_Subgroups = list(heatmap_centroid.keys())
            topk_elements = []            
            for lbl_n, lbl in enumerate(list(set(list_labels))):
                locate_perLabel = np.array([ttt==lbl for ttt in patientSubgroups[dataset.experiment_label[0]]])
                for TME_i in PIR_df.keys():
                    PIRs_for_one_TME = np.array([ttt == TME_i for ttt in patientSubgroups['Microenvironment element']])
                    PIRS_and_label = np.logical_and(PIRs_for_one_TME,locate_perLabel)
                    list_PIRs = [[n_t,t] for n_t, t in enumerate(patientSubgroups['PIR value']) if PIRS_and_label[n_t]]
                    list_PIRs.sort(key = lambda x: x[1])
                    for sorted_V in list_PIRs[:-topk:-1]:                        
                        topk_elements.append(sorted_V[0])
            selected_Patients = {'Patient subgroup':[], 'Patient index': [], 'PIR value':[], 'TME':[], 'TME -h':[],'Patient Name':[], 'Patient class':[]} 
            for p_ind in topk_elements:
                selected_Patients['Patient subgroup'].append(patientSubgroups['Patient subgroup'][p_ind])
                selected_Patients['Patient index'].append(patientSubgroups['Patient_index'][p_ind])
                selected_Patients['PIR value'].append(patientSubgroups['PIR value'][p_ind])
                TME_ind = listOfIndicesForTME[list(PIR_df.keys()).index(patientSubgroups['Microenvironment element'][p_ind])]
                selected_Patients['TME'].append([[i for i in Top1PerPatient if i[0]==TME_ind][0]])
                selected_Patients['TME -h'].append(patientSubgroups['Microenvironment element'][p_ind].split('*')[0])
                selected_Patients['Patient Name'].append(patientSubgroups['Patient_name'][p_ind])
                selected_Patients['Patient class'].append(patientSubgroups[dataset.experiment_label[0]][p_ind])
     

            # # TOP-2 PIRs for each TME.
            # topk = 2
            # ALLPatient_Subgroups = list(heatmap_centroid.keys())
            # topk_elements = []
            # for n_TME, PIRs_for_one_TME in enumerate(np.transpose(np.array([heatmap_centroid[PS] for PS in heatmap_centroid.keys()]))):
            #     for lbl_n, lbl in enumerate(list(set(list_labels))):
            #         locate_perLabel = np.array([lbl in hc.split(':')[1] for hc in list(heatmap_centroid.keys())])
            #         sorted_indices_PIR = PIRs_for_one_TME.argsort()[::-1]
            #         for sorted_V in sorted_indices_PIR:
            #             if locate_perLabel[sorted_V]:
            #                 topk_elements.append([ALLPatient_Subgroups[sorted_V],n_TME])
            #                 break        
            # Show PIRs mean values for patient subgroups
            # if len(heatmap_centroid[list(heatmap_centroid.keys())[0]])>1:
            #     plt.close('all')
            #     plt.figure()    
            #     df2 = pd.DataFrame.from_dict(heatmap_centroid).T        
            #     scm = sns.clustermap(df2,row_cluster=True,col_cluster=True,vmin=-2,vmax=2,xticklabels=TME_Names)
            #     plt.savefig(dataset.bioInsights_dir_abundance_analysis_Subgroups+'/Clustermap_SubgroupsDifferencesPIR.png',dpi=600)                                 
            #     # Show TME abundance differences for patient subroups highlighting top2 PIR values.
            #     plt.close('all')
            #     plt.figure()    
            #     df2 = pd.DataFrame.from_dict(heatmap_TME)        
            #     ax = sns.clustermap(df2,row_cluster=True,col_cluster=True,vmin=-2,vmax=2,yticklabels=list(heatmap_centroid.keys()))
            #     # from matplotlib.patches import Rectangle
            #     # ax = ax.ax_heatmap
            #     # for topkEL in topk_elements:            
            #     #     ax.add_patch(Rectangle((topkEL[1], ALLPatient_Subgroups.index(topkEL[0])), 1, 1, fill=False, edgecolor='blue',  lw=3))                
            #     plt.savefig(dataset.bioInsights_dir_abundance_analysis_Subgroups+'/Clustermap_SubgroupsDifferencesTME_iter_{}.png'.format(str(dataset.args['epochs'])),dpi=600)
            # Show PIR's differences between patient subgroups
            df = pd.DataFrame(dict([(k,pd.Series(v)) for k,v in patientSubgroups.items()])) 
            df.to_excel(dataset.bioInsights_dir_abundance_analysis_Subgroups+'Subgroup_Information_{}.xlsx'.format(str(dataset.args['epochs'])))
            # # Return pairs of patients that show mechanisms.
            # selected_Patients = {'Patient subgroup':[], 'Patient index': [], 'PIR value':[], 'TME':[], 'TME -h':[],'Patient Name':[], 'Patient class':[]} 
            # for topkEL in topk_elements[:np.array([t[0]<0.05 for t in top1ClustersStats]).sum()*2]:
            #     PS_Located = np.array(df['Patient subgroup']==int(topkEL[0].split('-PS')[1])-1)
            #     clss_Located = np.array(df[dataset.experiment_label[0]]==topkEL[0].split(':')[-1].split('-PS')[0]) 
            #     TME_Located = np.array(df['Microenvironment element']==TME_Names_Raw[topkEL[1]])
            #     Patient_PIR_max_Val = df['PIR value'][np.logical_and(np.logical_and(PS_Located,clss_Located),TME_Located)].max()
            #     PIR_max_val = np.array(df['PIR value']==Patient_PIR_max_Val)
            #     patient_indx = df['Patient_index'][np.logical_and(np.logical_and(PS_Located,clss_Located),TME_Located)].array
            #     for p_ind in patient_indx:
            #         selected_Patients['Patient subgroup'].append(int(topkEL[0].split('-PS')[1]))
            #         selected_Patients['Patient index'].append(p_ind)
            #         selected_Patients['PIR value'].append(Patient_PIR_max_Val)
            #         selected_Patients['TME'].append([[t for t in Top1PerPatient if t[0]==listOfIndicesForTME[topkEL[1]]][0][2]])
            #         selected_Patients['TME -h'].append(TME_Names_Raw[topkEL[1]])
            #         selected_Patients['Patient Name'].append(IndexAndClass[p_ind][0])
            #         selected_Patients['Patient class'].append(topkEL[0].split(':')[-1].split('-PS')[0])

            # # Intra class differences            
            # for clss in list(set(list_labels)):
            #     for micro_elem in set(patientSubgroups['Microenvironment element']):                    
            #         for subgrp_0 in set(np.array(patientSubgroups['Patient subgroup'])[np.array(patientSubgroups[dataset.experiment_label[0]])==clss]):
            #             for subgrp_1 in set(np.array(patientSubgroups['Patient subgroup'])[np.array(patientSubgroups[dataset.experiment_label[0]])==clss]):                        
            #                 clss_d = np.array(df[dataset.experiment_label[0]]==clss)
            #                 micro_elem_d = np.array(df['Microenvironment element']==micro_elem)
            #                 subgrp_0_d = np.array(df['Patient subgroup']==subgrp_0)
            #                 subgrp_1_d = np.array(df['Patient subgroup']==subgrp_1)
            #                 s1 = np.logical_and(np.logical_and(clss_d,micro_elem_d),subgrp_0_d)
            #                 s2 = np.logical_and(np.logical_and(clss_d,micro_elem_d),subgrp_1_d)                                                        
            #                 try:
            #                     p_val = mannwhitneyu(df['PIR value'][s1],df['PIR value'][s2])
            #                 except:
            #                     p_val=[1,1]
            #                 if p_val[1]<0.001:
            #                     bxplt = {'Patient subgroup':['PS'+str(subgrp_0)]*len(list(df['PIR value'][s1]))+['PS'+str(subgrp_1)]*len(list(df['PIR value'][s2])),'{} PIR values per patient'.format(micro_elem):list(df['PIR value'][s1])+list(df['PIR value'][s2])}
            #                     df_b = pd.DataFrame(dict([ (k,pd.Series(v)) for k,v in bxplt.items() ])) 
            #                     # Display and save boxplots of abundance.                
            #                     plt.close('all')
            #                     plt.figure()
            #                     ax = sns.boxplot(x='Patient subgroup', y='{} PIR values per patient'.format(micro_elem),data=df_b,order=['PS'+str(subgrp_0),'PS'+str(subgrp_1)])#, hue=dataset.experiment_label[0],data=df_b)                
            #                     test_results = add_stat_annotation(ax, data=df_b,x='Patient subgroup', y='{} PIR values per patient'.format(micro_elem),order=['PS'+str(subgrp_0),'PS'+str(subgrp_1)],
            #                                                     box_pairs=[('PS'+str(subgrp_0),'PS'+str(subgrp_1))],
            #                                                     test='Mann-Whitney', text_format='star',
            #                                                     loc='inside', verbose=2)                                                                                                                
            #                     plt.title('Intraclass Differences {} {} pval{}'.format(clss,micro_elem,np.round(p_val[1],4)))
            #                     plt.savefig(dataset.bioInsights_dir_abundance_analysis_Subgroups+'IntraClassDiff_{}_{}.png'.format(clss,micro_elem),dpi=600)
                                

            # # Inter class differences            
            # for micro_elem in set(patientSubgroups['Microenvironment element']):                    
            #     for subgrp_0 in set(np.array(patientSubgroups['Patient subgroup'])[np.array(patientSubgroups[dataset.experiment_label[0]])==clss]):
            #         for subgrp_1 in set(np.array(patientSubgroups['Patient subgroup'])[np.array(patientSubgroups[dataset.experiment_label[0]])==clss]):                        
            #             clss_d = np.array(df[dataset.experiment_label[0]]==clss)
            #             micro_elem_d = np.array(df['Microenvironment element']==micro_elem)
            #             subgrp_0_d = np.array(df['Patient subgroup']==subgrp_0)
            #             subgrp_1_d = np.array(df['Patient subgroup']==subgrp_1)
            #             s1 = np.logical_and(np.logical_and(clss_d,micro_elem_d),subgrp_0_d)
            #             s2 = np.logical_and(np.logical_and(clss_d,micro_elem_d),subgrp_1_d)                                                        
            #             p_val = mannwhitneyu(df['PIR value'][s1],df['PIR value'][s2])
            #             if p_val[1]<0.001:
            #                 bxplt = {'PS'+str(subgrp_0):df['PIR value'][s1],
            #                             'PS'+str(subgrp_1):df['PIR value'][s2]}
            #                 df_b = pd.DataFrame(dict([ (k,pd.Series(v)) for k,v in bxplt.items() ])) 
            #                 plt.close('all')
            #                 fig = plt.figure()
            #                 ax = fig.add_subplot(111)
            #                 ax.boxplot([np.array(df['PIR value'][s1]),np.array(df['PIR value'][s2])])
            #                 ax.set_xticklabels(['PS'+str(subgrp_0), 'PS'+str(subgrp_1)])                                
            #                 plt.title('IntraClassDiff_{}_{}_pval{}.png'.format(clss,micro_elem,np.round(p_val[1],4)))
            #                 plt.savefig(dataset.bioInsights_dir_abundance_analysis+'IntraClassDiff_{}_{}.png'.format(clss,micro_elem),dpi=600)

      
            # Show PIR's differences between patient subgroups
            df = pd.DataFrame(dict([(k,pd.Series(v)) for k,v in selected_Patients.items()])) 
            df.to_excel(dataset.bioInsights_dir_abundance_analysis_Subgroups+'Patients_toSHow_{}.xlsx'.format(str(dataset.args['epochs'])))

            # # Return best patient to show for each subgroup of patients.
            # Patient_subgroup = []
            # for clss in list(set(list_labels)):
            #     for subgrp_0 in set(np.array(patientSubgroups['Patient subgroup'])[np.array(patientSubgroups[dataset.experiment_label[0]])==clss]):                    
            #         subgrp_0_d = np.array(df['Patient subgroup']==subgrp_0)
            #         clss_d = np.array(df[dataset.experiment_label[0]]==clss)                    
            #         Patient_PIR_max_Val = df['PIR value'][np.logical_and(clss_d,subgrp_0_d)].max()
            #         Patient_subgroup.append(list(df['Patient_index'][df['PIR value']==Patient_PIR_max_Val])[0])

            print('BioInsights: Extraction of most important PIRs per TME')
            return selected_Patients, heatmap_centroid, Subgroups_of_patients_labels, Top1PerPatient

        def Initialize_model(dataset, heatmapALL,labels):
            '''
            '''
            # Initialize Variables            
            dependentVariable = np.concatenate(tuple([heatmapALL[h] for h in heatmapALL]),1)
            dependentVariable = (dependentVariable-dependentVariable.mean(0))/(dependentVariable.std(0)+1e-16)
            labels = np.array(labels[list(labels)[0]])

            # Initialize Model, and obtain unrestricted model.
            # model = Load_Model(dependentVariable.shape[1],labels.max()+1) 
            labels_names = [l[0] for l in labels] 
            labels_int = [list(set(labels_names)).index(l) for l in labels_names]       
            loadmodel_pls= False
            if loadmodel_pls:
                model = Load_Model(dataset)
                model.to('cpu')
            else:
                model = Model(dependentVariable.shape[1],len(set(labels_names)))             
                model.to('cpu')
            model.train()
            model, ACC, Unrestricted_loss = trainModel(model,dependentVariable,labels_int,loadmodel_pls)     
            ACC, Unrestricted_loss, pred_Confidence, real_class_confidence, pred_Label = validateModel(model, dependentVariable, labels_int)
            print('BioInsights: Unrestricted Model accuracy is ',ACC)
        
            return dependentVariable,model,labels_int,Unrestricted_loss,pred_Confidence, pred_Label, real_class_confidence
        
        def adjust_pvalue(Names,p_values,filename):
            Pvalue_adjust = {'Names':Names}
            Pvalue_adjust['lsu_values'] = lsu(np.array(p_values),alpha=0.05)[1]
            Pvalue_adjust['hochberg'] = hochberg(np.array(p_values))
            Pvalue_adjust['holm_bonferroni'] = holm_bonferroni(np.array(p_values))
            Pvalue_adjust['sidak'] = sidak(np.array(p_values))
            Pvalue_adjust['Original_pvalues'] = np.array(p_values)
            df = pd.DataFrame(Pvalue_adjust)
            df.to_excel(filename)  

        def TranslateIndexToCellTypes(dataset, topkClustersStats, heatmapALL, significantClusters):
        
            for topk in topkClustersStats:
                topk.append([])

            # Obtain the real number of the cluster
            statisticalTest = []
            index = 0
            for n,clusterLevel in enumerate(heatmapALL):
                for c in range(int(clusterLevel)):
                    if any([s==index for s in significantClusters]):
                        statisticalTest.append([pvalue_restrictedModel,clusterLevel,c])                    
                    for topk in topkClustersStats:
                        if topk[0]<0.05 and n==0 and c==0:
                            Top1PerPatient.append([topk[1],1])
                        if any([f==index for f in topk[1]]):
                            if n==0:
                                topk[4].append([pvalue_restrictedModel,'P'+str(c+1),c])
                            elif n==1:
                                topk[4].append([pvalue_restrictedModel,'N'+str(c+1),c])
                            elif n==2:
                                topk[4].append([pvalue_restrictedModel,'A'+str(c+1),c])    
                    for top1 in Top1PerPatient: 
                        if any([s==index for s in top1[0]]):           
                            top1.append([clusterLevel,c])
                    index+=1
            df = pd.DataFrame(topkClustersStats)
            df.to_excel(dataset.bioInsights_dir_abundance_analysis+"Significance_of_TMEs.xlsx")  
            print('BioInsights: Relevant TME/s '+' '.join([i[1] for i in topkClustersStats[0][4]])+' (p-val = '+str(round(topkClustersStats[0][4][0][0],4))+')')
            adjust_pvalue(Names=[topk[4] for topk in topkClustersStats],p_values=[topk[0] for topk in topkClustersStats],filename=dataset.bioInsights_dir_abundance_analysis+"Significance_of_TMEs_adjusted.xlsx")
            return Top1PerPatient, topkClustersStats, statisticalTest

        # Initialize regression model. 
        dependentVariable,model,labels_int,Unrestricted_loss,pred_Confidence, pred_Label,real_class_confidence = Initialize_model(dataset,heatmapALL,labels)

        # Find significant clusters using the model
        significantClusters, pvalue_restrictedModel, ACCRestricted, topkClustersStats, Top1PerPatient = findSignificantClusters(dataset,dependentVariable,model,labels_int,Unrestricted_loss,pred_Confidence,pvalueThreshold)        
        
        # Translate index to cell types.
        Top1PerPatient, topkClustersStats, significantClustersStats  = TranslateIndexToCellTypes(dataset,topkClustersStats, heatmapALL, significantClusters)                

        if not isTraining:
            # Visualize Regression analysis
            visualizeRegressionAnalysis(dataset,heatmapALL,labels,pred_Label,topkClustersStats,IndexAndClass,Top1PerPatient)

            # Obtain relevant PIRs
            topKClustersStats, RestrictedModelData = obtain_relevant_PIR(Top1PerPatient,topkClustersStats)

            # Calculate distinct groups of patients. To show intra-class heterogeneity            
            patient_Ineach_subgroup, Subgroups_of_patients_labels_names, Subgroups_of_patients_labels, Top1PerPatient = showGroups_Patients(RestrictedModelData,labels,dependentVariable,Top1PerPatient,topKClustersStats,dataset,IndexAndClass)

            # Show patient Prediction Confidence
            showPatientPredictionConfidence(dataset,pred_Confidence,labels,Subgroups_of_patients_labels_names, Subgroups_of_patients_labels,topKClustersStats,IndexAndClass,Top1PerPatient,real_class_confidence)            
        else:
            patient_Ineach_subgroup = []
        
        return topkClustersStats, Unrestricted_loss, Top1PerPatient, patient_Ineach_subgroup, real_class_confidence


import re
import copy
import logging

import torch
import torch_geometric
# from torch_sparse import coalesce, SparseTensor
# from torch_geometric.utils import (contains_isolated_nodes,
                                #    contains_self_loops, is_undirected)

from typing import Optional

import torch

def maybe_num_nodes(index: torch.Tensor,
                    num_nodes: Optional[int] = None) -> int:
    return int(index.max()) + 1 if num_nodes is None else num_nodes

__num_nodes_warn_msg__ = (
    'The number of nodes in your data object can only be inferred by its {} '
    'indices, and hence may result in unexpected batch-wise behavior, e.g., '
    'in case there exists isolated nodes. Please consider explicitly setting '
    'the number of nodes for this data object by assigning it to '
    'data.num_nodes.')


def size_repr(key, item, indent=0):
    indent_str = ' ' * indent
    if torch.is_tensor(item) and item.dim() == 0:
        out = item.item()
    elif torch.is_tensor(item):
        out = str(list(item.size()))
    # elif isinstance(item, SparseTensor):
    #     out = str(item.sizes())[:-1] + f', nnz={item.nnz()}]'
    elif isinstance(item, list) or isinstance(item, tuple):
        out = str([len(item)])
    elif isinstance(item, dict):
        lines = [indent_str + size_repr(k, v, 2) for k, v in item.items()]
        out = '{\n' + ',\n'.join(lines) + '\n' + indent_str + '}'
    elif isinstance(item, str):
        out = f'"{item}"'
    else:
        out = str(item)

    return f'{indent_str}{key}={out}'


class Data(object):
    r"""A plain old python object modeling a single graph with various
    (optional) attributes:

    Args:
        x (Tensor, optional): Node feature matrix with shape :obj:`[num_nodes,
            num_node_features]`. (default: :obj:`None`)
        edge_index (LongTensor, optional): Graph connectivity in COO format
            with shape :obj:`[2, num_edges]`. (default: :obj:`None`)
        edge_attr (Tensor, optional): Edge feature matrix with shape
            :obj:`[num_edges, num_edge_features]`. (default: :obj:`None`)
        y (Tensor, optional): Graph or node targets with arbitrary shape.
            (default: :obj:`None`)
        pos (Tensor, optional): Node position matrix with shape
            :obj:`[num_nodes, num_dimensions]`. (default: :obj:`None`)
        norm (Tensor, optional): Normal vector matrix with shape
            :obj:`[num_nodes, num_dimensions]`. (default: :obj:`None`)
        face (LongTensor, optional): Face adjacency matrix with shape
            :obj:`[3, num_faces]`. (default: :obj:`None`)

    The data object is not restricted to these attributes and can be extented
    by any other additional data.

    Example::

        data = Data(x=x, edge_index=edge_index)
        data.train_idx = torch.tensor([...], dtype=torch.long)
        data.test_mask = torch.tensor([...], dtype=torch.bool)
    """
    def __init__(self, x=None, edge_index=None, edge_attr=None, y=None,
                 pos=None, norm=None, face=None, **kwargs):
        self.x = x
        self.edge_index = edge_index
        self.edge_attr = edge_attr
        self.y = y
        self.pos = pos
        self.norm = norm
        self.face = face
        for key, item in kwargs.items():
            if key == 'num_nodes':
                self.__num_nodes__ = item
            else:
                self[key] = item
                
        if face is not None and face.dtype != torch.long:
            raise ValueError(
                (f'Argument `face` needs to be of type `torch.long` but found '
                 f'type `{face.dtype}`.'))

        if torch_geometric.is_debug_enabled():
            self.debug()

    @classmethod
    def from_dict(cls, dictionary):
        r"""Creates a data object from a python dictionary."""
        data = cls()

        for key, item in dictionary.items():
            data[key] = item

        if torch_geometric.is_debug_enabled():
            data.debug()

        return data

    def __getitem__(self, key):
        r"""Gets the data of the attribute :obj:`key`."""
        return getattr(self, key, None)

    def __setitem__(self, key, value):
        """Sets the attribute :obj:`key` to :obj:`value`."""
        setattr(self, key, value)

    @property
    def keys(self):
        r"""Returns all names of graph attributes."""
        keys = [key for key in self.__dict__.keys() if self[key] is not None]
        keys = [key for key in keys if key[:2] != '__' and key[-2:] != '__']
        return keys

    def __len__(self):
        r"""Returns the number of all present attributes."""
        return len(self.keys)

    def __contains__(self, key):
        r"""Returns :obj:`True`, if the attribute :obj:`key` is present in the
        data."""
        return key in self.keys

    def __iter__(self):
        r"""Iterates over all present attributes in the data, yielding their
        attribute names and content."""
        for key in sorted(self.keys):
            yield key, self[key]

    def __call__(self, *keys):
        r"""Iterates over all attributes :obj:`*keys` in the data, yielding
        their attribute names and content.
        If :obj:`*keys` is not given this method will iterative over all
        present attributes."""
        for key in sorted(self.keys) if not keys else keys:
            if key in self:
                yield key, self[key]

    def __cat_dim__(self, key, value):
        r"""Returns the dimension for which :obj:`value` of attribute
        :obj:`key` will get concatenated when creating batches.

        .. note::

            This method is for internal use only, and should only be overridden
            if the batch concatenation process is corrupted for a specific data
            attribute.
        """
        # Concatenate `*index*` and `*face*` attributes in the last dimension.
        if bool(re.search('(index|face)', key)):
            return -1
        # By default, concatenate sparse matrices diagonally.
        # elif isinstance(value, SparseTensor):
        #     return (0, 1)
        return 0

    def __inc__(self, key, value):
        r"""Returns the incremental count to cumulatively increase the value
        of the next attribute of :obj:`key` when creating batches.

        .. note::

            This method is for internal use only, and should only be overridden
            if the batch concatenation process is corrupted for a specific data
            attribute.
        """
        # Only `*index*` and `*face*` attributes should be cumulatively summed
        # up when creating batches.
        return self.num_nodes if bool(re.search('(index|face)', key)) else 0

    @property
    def num_nodes(self):
        r"""Returns or sets the number of nodes in the graph.

        .. note::
            The number of nodes in your data object is typically automatically
            inferred, *e.g.*, when node features :obj:`x` are present.
            In some cases however, a graph may only be given by its edge
            indices :obj:`edge_index`.
            PyTorch Geometric then *guesses* the number of nodes
            according to :obj:`edge_index.max().item() + 1`, but in case there
            exists isolated nodes, this number has not to be correct and can
            therefore result in unexpected batch-wise behavior.
            Thus, we recommend to set the number of nodes in your data object
            explicitly via :obj:`data.num_nodes = ...`.
            You will be given a warning that requests you to do so.
        """
        if hasattr(self, '__num_nodes__'):
            return self.__num_nodes__
        for key, item in self('x', 'pos', 'norm', 'batch'):
            return item.size(self.__cat_dim__(key, item))
        if hasattr(self, 'adj'):
            return self.adj.size(0)
        if hasattr(self, 'adj_t'):
            return self.adj_t.size(1)
        # if self.face is not None:
        #     logging.warning(__num_nodes_warn_msg__.format('face'))
        #     return maybe_num_nodes(self.face)
        # if self.edge_index is not None:
        #     logging.warning(__num_nodes_warn_msg__.format('edge'))
        #     return maybe_num_nodes(self.edge_index)
        return None

    @num_nodes.setter
    def num_nodes(self, num_nodes):
        self.__num_nodes__ = num_nodes

    @property
    def num_edges(self):
        r"""Returns the number of edges in the graph."""
        for key, item in self('edge_index', 'edge_attr'):
            return item.size(self.__cat_dim__(key, item))
        return None

    @property
    def num_faces(self):
        r"""Returns the number of faces in the mesh."""
        if self.face is not None:
            return self.face.size(self.__cat_dim__('face', self.face))
        return None

    @property
    def num_node_features(self):
        r"""Returns the number of features per node in the graph."""
        if self.x is None:
            return 0
        return 1 if self.x.dim() == 1 else self.x.size(1)

    @property
    def num_features(self):
        r"""Alias for :py:attr:`~num_node_features`."""
        return self.num_node_features

    @property
    def num_edge_features(self):
        r"""Returns the number of features per edge in the graph."""
        if self.edge_attr is None:
            return 0
        return 1 if self.edge_attr.dim() == 1 else self.edge_attr.size(1)

    def is_coalesced(self):
        r"""Returns :obj:`True`, if edge indices are ordered and do not contain
        duplicate entries."""
        edge_index, _ = coalesce(self.edge_index, None, self.num_nodes,
                                 self.num_nodes)
        return self.edge_index.numel() == edge_index.numel() and (
            self.edge_index != edge_index).sum().item() == 0

    def coalesce(self):
        r""""Orders and removes duplicated entries from edge indices."""
        self.edge_index, self.edge_attr = coalesce(self.edge_index,
                                                   self.edge_attr,
                                                   self.num_nodes,
                                                   self.num_nodes)
        return self

    # def contains_isolated_nodes(self):
    #     r"""Returns :obj:`True`, if the graph contains isolated nodes."""
    #     return contains_isolated_nodes(self.edge_index, self.num_nodes)

    # def contains_self_loops(self):
    #     """Returns :obj:`True`, if the graph contains self-loops."""
    #     return contains_self_loops(self.edge_index)

    # def is_undirected(self):
    #     r"""Returns :obj:`True`, if graph edges are undirected."""
    #     return is_undirected(self.edge_index, self.edge_attr, self.num_nodes)

    def is_directed(self):
        r"""Returns :obj:`True`, if graph edges are directed."""
        return not self.is_undirected()

    def __apply__(self, item, func):
        if torch.is_tensor(item):
            return func(item)
        # elif isinstance(item, SparseTensor):
        #     # Not all apply methods are supported for `SparseTensor`, e.g.,
        #     # `contiguous()`. We can get around it by capturing the exception.
        #     try:
        #         return func(item)
        #     except AttributeError:
        #         return item
        elif isinstance(item, (tuple, list)):
            return [self.__apply__(v, func) for v in item]
        elif isinstance(item, dict):
            return {k: self.__apply__(v, func) for k, v in item.items()}
        else:
            return item

    def apply(self, func, *keys):
        r"""Applies the function :obj:`func` to all tensor attributes
        :obj:`*keys`. If :obj:`*keys` is not given, :obj:`func` is applied to
        all present attributes.
        """
        for key, item in self(*keys):
            self[key] = self.__apply__(item, func)
        return self

    def contiguous(self, *keys):
        r"""Ensures a contiguous memory layout for all attributes :obj:`*keys`.
        If :obj:`*keys` is not given, all present attributes are ensured to
        have a contiguous memory layout."""
        return self.apply(lambda x: x.contiguous(), *keys)

    def to(self, device, *keys, **kwargs):
        r"""Performs tensor dtype and/or device conversion to all attributes
        :obj:`*keys`.
        If :obj:`*keys` is not given, the conversion is applied to all present
        attributes."""
        return self.apply(lambda x: x.to(device, **kwargs), *keys)

    def clone(self):
        return self.__class__.from_dict({
            k: v.clone() if torch.is_tensor(v) else copy.deepcopy(v)
            for k, v in self.__dict__.items()
        })

    def debug(self):
        if self.edge_index is not None:
            if self.edge_index.dtype != torch.long:
                raise RuntimeError(
                    ('Expected edge indices of dtype {}, but found dtype '
                     ' {}').format(torch.long, self.edge_index.dtype))

        if self.face is not None:
            if self.face.dtype != torch.long:
                raise RuntimeError(
                    ('Expected face indices of dtype {}, but found dtype '
                     ' {}').format(torch.long, self.face.dtype))

        if self.edge_index is not None:
            if self.edge_index.dim() != 2 or self.edge_index.size(0) != 2:
                raise RuntimeError(
                    ('Edge indices should have shape [2, num_edges] but found'
                     ' shape {}').format(self.edge_index.size()))

        if self.edge_index is not None and self.num_nodes is not None:
            if self.edge_index.numel() > 0:
                min_index = self.edge_index.min()
                max_index = self.edge_index.max()
            else:
                min_index = max_index = 0
            if min_index < 0 or max_index > self.num_nodes - 1:
                raise RuntimeError(
                    ('Edge indices must lay in the interval [0, {}]'
                     ' but found them in the interval [{}, {}]').format(
                         self.num_nodes - 1, min_index, max_index))

        if self.face is not None:
            if self.face.dim() != 2 or self.face.size(0) != 3:
                raise RuntimeError(
                    ('Face indices should have shape [3, num_faces] but found'
                     ' shape {}').format(self.face.size()))

        if self.face is not None and self.num_nodes is not None:
            if self.face.numel() > 0:
                min_index = self.face.min()
                max_index = self.face.max()
            else:
                min_index = max_index = 0
            if min_index < 0 or max_index > self.num_nodes - 1:
                raise RuntimeError(
                    ('Face indices must lay in the interval [0, {}]'
                     ' but found them in the interval [{}, {}]').format(
                         self.num_nodes - 1, min_index, max_index))

        if self.edge_index is not None and self.edge_attr is not None:
            if self.edge_index.size(1) != self.edge_attr.size(0):
                raise RuntimeError(
                    ('Edge indices and edge attributes hold a differing '
                     'number of edges, found {} and {}').format(
                         self.edge_index.size(), self.edge_attr.size()))

        if self.x is not None and self.num_nodes is not None:
            if self.x.size(0) != self.num_nodes:
                raise RuntimeError(
                    ('Node features should hold {} elements in the first '
                     'dimension but found {}').format(self.num_nodes,
                                                      self.x.size(0)))

        if self.pos is not None and self.num_nodes is not None:
            if self.pos.size(0) != self.num_nodes:
                raise RuntimeError(
                    ('Node positions should hold {} elements in the first '
                     'dimension but found {}').format(self.num_nodes,
                                                      self.pos.size(0)))

        if self.norm is not None and self.num_nodes is not None:
            if self.norm.size(0) != self.num_nodes:
                raise RuntimeError(
                    ('Node normals should hold {} elements in the first '
                     'dimension but found {}').format(self.num_nodes,
                                                      self.norm.size(0)))

    def __repr__(self):
        cls = str(self.__class__.__name__)
        has_dict = any([isinstance(item, dict) for _, item in self])

        if not has_dict:
            info = [size_repr(key, item) for key, item in self]
            return '{}({})'.format(cls, ', '.join(info))
        else:
            info = [size_repr(key, item, indent=2) for key, item in self]
            return '{}(\n{}\n)'.format(cls, ',\n'.join(info))



import numpy as np
import os.path as osp
import copy
import os
import matplotlib.pyplot as plt
import cv2
from skimage.morphology import opening, disk
from skimage.measure import perimeter
from matplotlib import cm
from scipy import stats
import seaborn as sns
from torch import clamp_min
from NaroNet.utils.parallel_process import parallel_process
from tifffile.tifffile import imwrite
import pandas as pd
import gc
import itertools
from skimage import filters
import time 
from openTSNE import TSNE
from sklearn.preprocessing import StandardScaler
from torch_geometric.data.makedirs import makedirs


def visualize_tsne_phenotypes(dataset, IndexAndClass,clusters):

    def load_embeddings(patch_dir):
        
        # Initialize features    
        features = []
        features_idx = []

        # Extract patches and save it into a graph
        for subject in IndexAndClass:                                    
            
            # Load image 
            subject_feats = np.load(patch_dir + subject[0] + '.npy')

            # Store feats
            features.append(subject_feats)
            
            # Store subject idx
            features_idx.append([subject[1]]*subject_feats.shape[0])

        # Store in one file all features
        features = np.concatenate(features,axis=0)
        features_idx = np.concatenate(features_idx,axis=0)    
        return features, features_idx
    
    def tsne_with_pheno_thrshold(dataset,features_idx,features,pheno_assignmnents):
        
        # Select number of patches
        num_val =min(10000000,features_idx.shape[0])     
        features_idx = features_idx[::int(features_idx.shape[0]/num_val)]
        features = features[::int(features.shape[0]/num_val),:]
        pheno_assignmnents = pheno_assignmnents[::int(features.shape[0]/num_val),:]
        
        # Reorder samples
        sel_vals = np.random.randint(0,high=features.shape[0],size=features.shape[0])
        features = features[sel_vals,:]
        features_idx = features_idx[sel_vals]
        pheno_assignmnents = pheno_assignmnents[sel_vals,:]
        
        # Normalize features
        scaled_features = StandardScaler().fit_transform(features[:,2:])

        # Create UMAP
        embedding = TSNE(n_jobs=8,random_state=3, verbose=True).fit(scaled_features)

        # Show each phenotype
        for ph in range(pheno_assignmnents.shape[1]):

            # Create and save scatter plot figure
            plt.close('all')
            fig, ax = plt.subplots()   
            scatter = ax.scatter(embedding[:, 0],embedding[:, 1],s=0.5,c=pheno_assignmnents[:,ph],vmin=0,vmax=1,cmap='jet',edgecolors='none')        
            plt.title('TSNE of the dataset Phenoptype - '+str(ph+1), fontsize=24)    
            plt.colorbar(scatter)
            plt.savefig(dataset.bioInsights_dir_cell_types_Pheno+'tSNE/'+'Ph_'+str(ph+1)+'.png',dpi=600)

        pheno_assignmnents_75 = np.zeros(pheno_assignmnents.shape[0])
        for ph in range(pheno_assignmnents.shape[1]):
            thr = np.percentile(pheno_assignmnents[pheno_assignmnents.argmax(1)==ph,ph],90)
            pheno_assignmnents_75[np.logical_and(pheno_assignmnents.argmax(1)==ph,pheno_assignmnents[:,ph]>=thr)] = ph+1
        plt.close('all')
        fig, ax = plt.subplots()   
        scatter = ax.scatter(embedding[:, 0],embedding[:, 1],s=0.5,c=pheno_assignmnents_75,cmap='jet',edgecolors='none')        
        plt.title('TSNE of the dataset ALL_Phenotypes D1', fontsize=24)    
        plt.colorbar(scatter)
        plt.savefig(dataset.bioInsights_dir_cell_types_Pheno+'tSNE/'+'ALL_Phenotypes_D1.png',dpi=600)        


    pat_num = int(len(IndexAndClass)*dataset.args['TSNE_Perc_Pat'])
    IndexAndClass = IndexAndClass[:pat_num]

    # Directory to load
    patch_dir = dataset.root+'Patch_Contrastive_Learning/Image_Patch_Representation/'       

    # Load data
    features, features_idx = load_embeddings(patch_dir)

    # Load pheno assignments
    pheno_assignmnents = []
    for subject_idx, subject_info in enumerate(IndexAndClass):              
        _, cell_type_assignment = load_cell_types_assignments(dataset, 0, subject_info, 0, clusters[0], [])
        pheno_assignmnents.append(cell_type_assignment[:(features_idx==subject_info[1]).sum(),:])
    pheno_assignmnents = np.concatenate(pheno_assignmnents,axis=0)   
        
    # Create Tsne 
    makedirs(dataset.bioInsights_dir_cell_types_Pheno+'tSNE/')
    embedding = tsne_with_pheno_thrshold(dataset,features_idx,features,pheno_assignmnents)


def load_cell_types_assignments(dataset, cell_type_idx, subject_info,subgraph_idx,n_cell_types, prev_cell_type_assignment):
    """
    Obtain matrix that assigns patches to cell types (phenotypes and neighborhoods)
    dataset: (object)
    cell_type_idx: (int) cell_type_idx==0 is phenotype, cell_type_idx==1 is neihborhood, cell_type_idx==2 is neihborhood interaction
    subject_info: (list of str and int)
    subgraph_idx: (int) specifying the index of the subgraph.
    n_cell_types: (int) number of cell types (phenotypes or neighborhoods)
    prev_cell_type_assignment: (array of int) specifying assignments
    """

    # If phenotype or neighborhood load matrix assignment.
    if cell_type_idx<2:
        cell_type_assignment = np.load(osp.join(dataset.processed_dir_cell_types,
                    'cluster_assignmentPerPatch_Index_{}_{}_ClustLvl_{}.npy'.format(subject_info[1], subgraph_idx, n_cell_types)))                        

    # if neighborhood interaction the matrix assigns neighbors to neighborhood interactions.
    else:        
        # Load neighborhood interactions
        secondorder_assignment = np.load(osp.join(dataset.processed_dir_cell_types,
                                'cluster_assignment_Index_{}_ClustLvl_{}.npy'.format(subject_info[1], n_cell_types)))                    
        
        # Obtain assignment of patches to neighborhood interactions
        cell_type_assignment = np.matmul(prev_cell_type_assignment,secondorder_assignment)
    prev_cell_type_assignment = copy.deepcopy(cell_type_assignment)
    
    return prev_cell_type_assignment, cell_type_assignment

def load_patch_image(dataset,imList):
    '''
    Load patch image
    # Create label map. 
    labels = []
    last_patch = 0
    for im_i in imList:
        num_patches = int(im_i.shape[0]/dataset.patch_size)*int(im_i.shape[1]/dataset.patch_size)
        labels.append(np.reshape(np.array(range(last_patch,last_patch+num_patches)),(int(im_i.shape[0]/dataset.patch_size),int(im_i.shape[1]/dataset.patch_size))))
        last_patch = last_patch+num_patches
    labels = np.concatenate(labels,1)
    '''             

    labels = []
    last_patch = 0
    x_y_shape = [0,0]
    for im_i in imList:

        # Create label image
        num_patches = int(im_i.shape[0]/dataset.patch_size)*int(im_i.shape[1]/dataset.patch_size)
        labels.append(np.transpose(np.reshape(np.array(range(last_patch,last_patch+num_patches)),(int(im_i.shape[1]/dataset.patch_size),int(im_i.shape[0]/dataset.patch_size)))))
        last_patch = last_patch+num_patches

        # Update shape of mosaic
        x_y_shape[0], x_y_shape[1] = max(labels[-1].shape[0],x_y_shape[0]), labels[-1].shape[1]+x_y_shape[1]
            
    # Create mosaic
    mosaic = np.zeros(x_y_shape,dtype=np.uint32)+4294967295

    # Concatenate images
    col=0
    for lab in labels:
        mosaic[0:lab.shape[0],col:col+lab.shape[1]] = lab
        col += lab.shape[1]

    return mosaic, labels

def select_patches_from_cohort_(clusters,dataset,subject_info,count):

    # Crops,Confidence,and Phenotype Vector.
    CropConfPheno = []
    CropConfTissueComm = []
    TissueArea_Conf = []
    for c in range(clusters[0]):
        CropConfPheno.append([])
            
    for c in range(clusters[1]):
        CropConfTissueComm.append([])  
    
    for c in range(clusters[2]):
        TissueArea_Conf.append([]) 

    # Crops,Confidence,and Phenotype Vector.    
    prev_cell_type_assignment=[]
        
    # Apply mask to patch
    for subgraph_idx in range(dataset.findLastIndex(subject_info[1])+1):                
        
        # Open Raw Image.        
        im, imList = dataset.open_Raw_Image(subject_info,1)                

        for cell_type_idx, n_cell_types in enumerate(clusters):    
            # load cell_types_assignments
            prev_cell_type_assignment, cell_type_assignment = load_cell_types_assignments(
                                                                    dataset,cell_type_idx,subject_info,subgraph_idx,
                                                                    n_cell_types,prev_cell_type_assignment)                               
            # Open Single-Cell Contrastive Learning Information
            PCL_reprsntions = np.load(dataset.raw_dir+'/{}.npy'.format(subject_info[0]))                     

            # Phenotypes and neighborhoods, or areas
            if cell_type_idx<2:             
                # Select patches with most confidence
                CropConfPheno,CropConfTissueComm = topk_confident_patches(dataset,n_cell_types,
                                                                            cell_type_assignment,cell_type_idx,CropConfPheno,
                                                                            CropConfTissueComm,im,imList,PCL_reprsntions,count)  
            else:
                # Save subject index and max cell type
                for c_type in range(n_cell_types):
                    TissueArea_Conf[c_type].append(cell_type_assignment[:,c_type])

    del im, imList, PCL_reprsntions
    gc.collect(0)
    gc.collect(1)
    gc.collect(2)
    time.sleep(3)

    return CropConfPheno,CropConfTissueComm,TissueArea_Conf


def select_patches_from_cohort(dataset,IndexAndClass,clusters):
    '''
    '''
     
    # Prepare parallel process
    dict_subjects = []
    for count, subject_info in enumerate(IndexAndClass):
        if subject_info[2][0]!='None':
            dict_subjects.append({'clusters':clusters,'dataset':dataset,'subject_info':subject_info,'count':count})
    
    # select_patches_from_cohort
    result_images = parallel_process(dict_subjects,select_patches_from_cohort_,use_kwargs=True,front_num=1,desc='BioInsights: Get relevant examples of cell types') 

    # Crops,Confidence,and Phenotype Vector.
    CropConfPheno = []
    CropConfTissueComm = []
    TissueArea_Conf = []
    for c in range(clusters[0]):
        CropConfPheno.append([])
            
    for c in range(clusters[1]):
        CropConfTissueComm.append([])  
    
    for c in range(clusters[2]):
        TissueArea_Conf.append([])  

    # Get lists of patches
    for cluster_level in result_images:  
        for clust_idx, cluster in enumerate(cluster_level[0]):
            for patch in cluster:  
                CropConfPheno[clust_idx].append(patch)
        for clust_idx, cluster in enumerate(cluster_level[1]):
            for patch in cluster:  
                CropConfTissueComm[clust_idx].append(patch)
        for r_i, r in enumerate(cluster_level[2]):
            TissueArea_Conf[r_i].append(r)
            
    return CropConfPheno,CropConfTissueComm, TissueArea_Conf

def topk_confident_patches(dataset,n_cell_types,
                            cell_type_assignment,cell_type_idx,CropConfPheno,
                            CropConfTissueComm,im,imList,PCL_reprsntions,count):
    '''
    '''
    K = int(np.ceil(10/dataset.processed_file_names*100))

    # Create label map. 
    labels,labels_list = load_patch_image(dataset,imList)
    # for ll in range(len(labels_list)):
    #     labels_list[ll] = np.transpose(labels_list[ll])
    # labels = np.concatenate(labels_list,axis=1)

            
    for c in range(n_cell_types):
                                                                        
        # Obtain patch info of the K with most certainty
        highest_ent_patches = cell_type_assignment[:PCL_reprsntions.shape[0],c].argsort()[-K:]
        for patch_idx in highest_ent_patches:                  
                    
            # Select the patch from the patch_image            
            mask = labels==patch_idx

            if cell_type_idx==0:
                mx_val = mask.argmax(0).max()*dataset.patch_size
                mxx_val = mask.argmax(1).max()*dataset.patch_size
                CropConfPheno[c].append([im[mx_val:mx_val+dataset.patch_size,mxx_val:mxx_val+dataset.patch_size].copy(), # The original Image
                                            cell_type_assignment[patch_idx,c], # The cell type certainty
                                            [], # The parameters obtained from contrastive learning
                                            [100000*count+patch_idx]]) # Number of the image, and patch identificator
            elif cell_type_idx==1:
                mx_val = mask.argmax(0).max()*dataset.patch_size
                mxx_val = mask.argmax(1).max()*dataset.patch_size
                minIdx = mx_val-dataset.patch_size*2
                maxIdx = mx_val+dataset.patch_size*3
                minIdy = mxx_val-dataset.patch_size*2
                maxIdy = mxx_val+dataset.patch_size*3
                if maxIdx>im.shape[0] or maxIdy>im.shape[1] or minIdx<0 or minIdy<0:
                    continue
                CropConfTissueComm[c].append([im[minIdx:maxIdx,minIdy:maxIdy].copy(), # The original Image
                                            cell_type_assignment[patch_idx,c], # The cell type certainty
                                            [100000*count+patch_idx]]) # Patch identificator 
    
    return CropConfPheno,CropConfTissueComm

def save_2Dmatrix_in_excel_with_names(filename,matrix,Names):
    dict_ = {}
    for n, name in enumerate(Names):
        dict_[name] = matrix[:,n]
    dict_ = pd.DataFrame.from_dict(dict_)      
    dict_.to_excel(filename) 

def save_heatmap_with_names(filename,matrix,Names):   
    import sys
    sys.setrecursionlimit(1000000) 
    if matrix.shape[0]>2:
        plt.close('all')
        plt.figure()
        sns.set(font_scale=1.1)
        scaler = StandardScaler()
        scaler.fit(matrix)        
        h_E_Fig = sns.clustermap(scaler.transform(matrix)*25+128, row_cluster=True, xticklabels=Names, linewidths=0,vmin=-2, vmax=2, cmap="bwr")                
        h_E_Fig.savefig(filename,dpi=600) 

def calculate_pearson_correlation(matrix_0,matrix_1):
    cor_val = stats.spearmanr(matrix_0.flatten(),matrix_1.flatten()).correlation
    cor_val = 0 if np.isnan(cor_val) else cor_val
    return cor_val

def calculate_marker_colocalization(matrix,MarkerNames):
    # matrix: contains (number of patches, x_dimension, y_dimension, number of markers)
    Marker_Colocalization =  np.ones((len(MarkerNames),len(MarkerNames)))
    for n_comb, pair_of_markers in enumerate(itertools.combinations(MarkerNames,2)):
        id_0 = MarkerNames.index(pair_of_markers[0])
        id_1 = MarkerNames.index(pair_of_markers[1])
        matrix_0 = matrix[:,:,:,[id_0]]
        matrix_1 = matrix[:,:,:,[id_1]]
        # for n_patch in range(matrix.shape[0]):            
        Marker_Colocalization[id_0,id_1] = calculate_pearson_correlation(matrix_0,matrix_1)
        Marker_Colocalization[id_1,id_0] = Marker_Colocalization[id_0,id_1]
    return Marker_Colocalization

def calculate_Cell_Size_circularity(filename, matrix):
    # matrix: contains (number of patches, x_dimension, y_dimension)         
    # selem = disk(2)
    Size = np.zeros((matrix.shape[3]))
    Eccentricity = np.zeros((matrix.shape[3]))
    Granularity = np.zeros((matrix.shape[3]))
    for mk in range(matrix.shape[3]):

        # Calculate global threshold
        thr = filters.threshold_otsu(matrix[:,:,:,mk])

        # Initialize count
        count = 0 

        for patch in range(matrix.shape[0]):
            
            if matrix[patch,:,:].sum()>0 and (matrix[patch,:,:].max()-matrix[patch,:,:].min())>0:
                logical = matrix[patch,:,:,mk]>thr
                output = cv2.connectedComponentsWithStats(np.array(logical,dtype=np.uint8), 4, cv2.CV_32S)

                if output[0]>1: # at least one component
                    Size[mk] += output[2][1:,4].mean()                    
                    Eccentricity[mk] += min((4*3.14159*(output[1]==1).sum())/(perimeter((output[1]==1),neighbourhood=4)**2),1) # Eccentricity
                    Granularity[mk] += output[0]-1

                    count += 1
        if count>0:
            Size[mk] /= count
            Eccentricity[mk] /= count
            Granularity[mk] /= count
        else:
            Size[mk] = 0
            Eccentricity[mk] = 0
            Granularity[mk] = 0

    return Size, Eccentricity, Granularity

def extract_topk_patches_from_cohort(dataset, CropConf, Marker_Names,cell_type,Thresholds):
    '''
    docstring
    '''

    thisfolder = dataset.bioInsights_dir_cell_types + cell_type+'/'

    if cell_type=='Phenotypes':
        mult_1 = 1 
        mult_2 = 2
    else:
        mult_1 = 5 
        mult_2 = 4

    ## Iterate through Phenotypes to extract topk patches
    k=100
    topkPatches=[]
    # Create a heatmap marker using topk patches
    heatmapMarkerExpression = np.zeros((len(CropConf),len(Marker_Names))) # Number of TMEs x number of markers
    heatmap_Colocalization = np.zeros((len(CropConf),len(Marker_Names),len(Marker_Names))) # Number of TMEs x Number of Markers x Number of Markers
    heatmap_Size = np.zeros((len(CropConf),len(Marker_Names))) # size and circularity of markers
    heatmap_Eccentricity = np.zeros((len(CropConf),len(Marker_Names))) # size and circularity of markers
    heatmap_Granularity = np.zeros((len(CropConf),len(Marker_Names))) # size and circularity of markers

    # Make directory
    makedirs(thisfolder+'Prototypes/')

    # Use CropCOnf, that saves a lot of patches...
    for n_cell_type ,CropConf_i in enumerate(CropConf):
        
        if len(CropConf_i)==0: # Let te heatmap be zero in case no cells were assigned to this TME.
            continue
        
        # Choose patches with most confidence
        topkPheno = np.array([CCP[1] for CCP in CropConf_i]).argsort()[-k:]               
        
        heatmap_Colocalization[n_cell_type,:,:] = calculate_marker_colocalization(np.array([CropConf_i[t][0] for t in topkPheno]),Marker_Names)    
        heatmap_Size[n_cell_type,:], heatmap_Eccentricity[n_cell_type,:], heatmap_Granularity[n_cell_type,:] = calculate_Cell_Size_circularity(thisfolder,np.array([CropConf_i[t][0] for t in topkPheno]))

        # Save topkPheno to heatMarkerMap. Mean
        MarkerExpression = np.array([CropConf_i[t][0].mean((0,1)) for t in topkPheno])
              
        heatmapMarkerExpression[n_cell_type,:] = np.mean(MarkerExpression,axis=0)                    
      
        # Save Image in RGB
        ImwithKPatches = np.zeros((dataset.patch_size*mult_1*int(np.sqrt(k))+mult_2*int(np.sqrt(k)),dataset.patch_size*mult_1*int(np.sqrt(k))+mult_2*int(np.sqrt(k)),CropConf_i[0][0].shape[2]))
        for t_n, t in enumerate(topkPheno):
            row = np.floor(t_n/int(k**0.5))
            col = np.mod(t_n,int(k**0.5))
       
            # Assign patch to Image
            ImwithKPatches[int(row*mult_1*dataset.patch_size+row*mult_2):int((row+1)*mult_1*dataset.patch_size+row*mult_2),int(col*mult_1*dataset.patch_size+col*mult_2):int((col+1)*mult_1*dataset.patch_size+col*mult_2),:] = CropConf_i[t][0]
        
        # Fill unassigned patches with zeroes.
        for t_n in range(len(topkPheno),k):
            row = np.floor(t_n/int(k**0.5))
            col = np.mod(t_n,int(k**0.5))
            # Assign patch to Image
            ImwithKPatches[int(row*mult_1*dataset.patch_size+row*mult_2):int((row+1)*mult_1*dataset.patch_size+row*mult_2),int(col*mult_1*dataset.patch_size+col*mult_2):int((col+1)*mult_1*dataset.patch_size+col*mult_2),:] = 0.0
                        
        # RGBImwithKPatches = dataset.nPlex2RGB(ImwithKPatches)
        imwrite(thisfolder+'Prototypes/Cell_type_{}_Raw.tiff'.format(n_cell_type+1),np.moveaxis(ImwithKPatches,2,0), photometric='minisblack')
        
        # Save Certainty of this Phenotype
        plt.close('all')
        plt.figure()
        n, bins, patches = plt.hist(np.array([i[1] for i in CropConf_i]),range=(0, 1), color=cm.jet_r(int(n_cell_type*(255/int(len(CropConf))))), alpha=1)            
        plt.ylabel('Number of Patches',fontsize=16)            
        plt.xlabel('Level of TME certainty',fontsize=16)
        plt.title('Histogram of TME ' + str(n_cell_type+1) + ' certainty',fontsize=16)
        plt.savefig(thisfolder+'Prototypes/ConfidenceHistogram_{}.png'.format(n_cell_type+1), format="PNG",dpi=600)                                         

        # Assign topkPheno Patches
        topkPatches+=[CropConf_i[t] for t in topkPheno]     

    return heatmapMarkerExpression, heatmap_Colocalization, heatmap_Size, heatmap_Eccentricity, heatmap_Granularity

def save_heatmap_raw_and_normalized(filename, heatmap, TME_names,Colormap,Marker_Names):    
    y_ticklabels = TME_names[heatmap.sum(1)!=0]
    c_map = Colormap[:heatmap.shape[0]][heatmap.sum(1)!=0]
    # Figure Heatmap Raw Values
    plt.close('all')
    plt.figure()
    sns.set(font_scale=0.9)
    cluster=True if (heatmap.sum(1)!=0).sum()>1 else False
    h_E_Fig = sns.clustermap(heatmap[heatmap.sum(1)!=0,:],col_cluster=cluster, row_cluster=cluster, row_colors=c_map,xticklabels=Marker_Names,yticklabels=y_ticklabels,vmin=-1,vmax=1, linewidths=0.5, cmap="bwr")            
    h_E_Fig.savefig(filename+'_Raw.png',dpi=600) 

    # Save Excel.
    df = pd.DataFrame(heatmap[heatmap.sum(1)!=0,:], columns = Marker_Names)
    df.index = y_ticklabels
    df.to_excel(filename+'_Raw.xlsx')  

    # Figure Heatmap Min is 0 and max is 1 Values    
    h_E_COL_MinMax = heatmap[heatmap.sum(1)!=0,:] - heatmap[heatmap.sum(1)!=0,:].min(0,keepdims=True)
    h_E_COL_MinMax = h_E_COL_MinMax/h_E_COL_MinMax.max(0,keepdims=True)
    h_E_COL_MinMax[np.isnan(h_E_COL_MinMax)] = 0 
    plt.close('all')
    plt.figure()
    sns.set(font_scale=0.9)
    if (h_E_COL_MinMax.sum(1)!=0).sum()==1: # One column heatmap
        h_E_Fig = sns.clustermap(h_E_COL_MinMax[h_E_COL_MinMax.sum(1)!=0,:],col_cluster=False, 
            row_cluster=False, row_colors=c_map,xticklabels=Marker_Names,yticklabels=y_ticklabels, linewidths=0.5,vmin=-1,vmax=1, cmap="bwr")            
        h_E_Fig.savefig(filename+'_MinMax.png',dpi=600) 
    elif (h_E_COL_MinMax.sum(1)!=0).sum()>1: # Normal heatmap.    
        h_E_Fig = sns.clustermap(h_E_COL_MinMax[h_E_COL_MinMax.sum(1)!=0,:],col_cluster=True, 
            row_cluster=True, row_colors=c_map,xticklabels=Marker_Names,yticklabels=y_ticklabels, linewidths=0.5,vmin=-1,vmax=1, cmap="bwr")            
        h_E_Fig.savefig(filename+'_MinMax.png',dpi=600)  
    
    # Figure Heatmap z-scored values
    h_E_COL_Norm = stats.zscore(heatmap[heatmap.sum(1)!=0,:],axis=0)  
    h_E_COL_Norm[np.isnan(h_E_COL_Norm)] = 0              
    plt.close('all')
    plt.figure()
    sns.set(font_scale=0.9)    
    if (h_E_COL_Norm.sum(1)!=0).sum()==1: # One column heatmap
        h_E_Fig = sns.clustermap(h_E_COL_Norm,col_cluster=False, vmin=-1, vmax=1, row_cluster=False, row_colors=c_map,xticklabels=Marker_Names,yticklabels=y_ticklabels, linewidths=0.5, cmap="bwr")            
        h_E_Fig.savefig(filename+'_Norm.png',dpi=600) 
    elif (h_E_COL_Norm.sum(1)!=0).sum()>1: # Normal heatmap
        h_E_Fig = sns.clustermap(h_E_COL_Norm,col_cluster=True, vmin=-1, vmax=1, row_cluster=True, row_colors=c_map,xticklabels=Marker_Names,yticklabels=y_ticklabels, linewidths=0.5, cmap="bwr")            
        h_E_Fig.savefig(filename+'_Norm.png',dpi=600) 

    # Figure Heatmap z-scored values
    h_E_COL_Norm = stats.zscore(h_E_COL_Norm[heatmap.sum(1)!=0,:],axis=1)  
    h_E_COL_Norm[np.isnan(h_E_COL_Norm)] = 0              
    plt.close('all')
    plt.figure()
    sns.set(font_scale=0.9)    
    if (h_E_COL_Norm.sum(1)!=0).sum()==1: # One column heatmap
        h_E_Fig = sns.clustermap(h_E_COL_Norm,col_cluster=False, vmin=-1, vmax=1, row_cluster=False, row_colors=c_map,xticklabels=Marker_Names,yticklabels=y_ticklabels, linewidths=0.5, cmap="bwr")            
        h_E_Fig.savefig(filename+'_Norm_rows.png',dpi=600) 
    elif (h_E_COL_Norm.sum(1)!=0).sum()>1: # Normal heatmap
        h_E_Fig = sns.clustermap(h_E_COL_Norm,col_cluster=True, vmin=-1, vmax=1, row_cluster=True, row_colors=c_map,xticklabels=Marker_Names,yticklabels=y_ticklabels, linewidths=0.5, cmap="bwr")            
        h_E_Fig.savefig(filename+'_Norm_rows.png',dpi=600) 

def save_heatMapMarker_and_barplot(dataset, heatmapMarkerExpression, heatmap_Colocalization, heatmap_Size, heatmap_Eccentricity, heatmap_Granularity, CropConf,Marker_Names,cell_type,Thresholds):
    '''
    '''
    if cell_type=='Phenotypes':
        abrev = 'P'
        names = np.array([abrev+str(i+1) for i in range(len(CropConf))])

        # Exclude phenotypes from the analysis
        phnos = np.array(range(heatmapMarkerExpression.shape[0]))
        for excluded in dataset.args['Pheno_Excluded_from_analysis']:            
            phnos = phnos[phnos!=excluded-1]
        heatmapMarkerExpression = heatmapMarkerExpression[phnos,:]
        heatmap_Size = heatmap_Size[phnos,:]
        heatmap_Eccentricity = heatmap_Eccentricity[phnos,:]
        heatmap_Granularity = heatmap_Granularity[phnos,:]
        names = names[phnos]
        
    else:
        abrev = 'N'
        names = np.array([abrev+str(i+1) for i in range(len(CropConf))])

        # Exclude phenotypes from the analysis
        neighs = np.array(range(heatmapMarkerExpression.shape[0]))
        for excluded in dataset.args['Neigh_Excluded_from_analysis']:            
            neighs = neighs[neighs!=excluded-1]
        heatmapMarkerExpression = heatmapMarkerExpression[neighs,:]
        heatmap_Size = heatmap_Size[neighs,:]
        heatmap_Eccentricity = heatmap_Eccentricity[neighs,:]
        heatmap_Granularity = heatmap_Granularity[neighs,:]
        names = names[neighs]

    Colormap=cm.jet(range(0,255,int(255/heatmapMarkerExpression.shape[0])))[:,:3]

    # Save heatmapmarker to disk    
    makedirs(dataset.bioInsights_dir_cell_types+cell_type+'/Marker_Expression/')
    save_heatmap_raw_and_normalized(dataset.bioInsights_dir_cell_types+cell_type+'/Marker_Expression/heatmap', 
                                        heatmapMarkerExpression, names,
                                        Colormap,Marker_Names)

    makedirs(dataset.bioInsights_dir_cell_types+cell_type+'/Morphology/')
    save_heatmap_raw_and_normalized(dataset.bioInsights_dir_cell_types+cell_type+'/Morphology/Heatmap_Size', 
                                        heatmap_Size, names,
                                        Colormap,Marker_Names)            

    save_heatmap_raw_and_normalized(dataset.bioInsights_dir_cell_types+cell_type+'/Morphology/heatmap_Eccentricity', 
                                        heatmap_Eccentricity, names,
                                        Colormap,Marker_Names)         

    save_heatmap_raw_and_normalized(dataset.bioInsights_dir_cell_types+cell_type+'/Morphology/heatmap_Granularity', 
                                        heatmap_Granularity, names,
                                        Colormap,Marker_Names)     

    # Correlation matrix
    makedirs(dataset.bioInsights_dir_cell_types+cell_type+'/Marker_Colocalization/')
    mask = np.triu(np.ones_like(heatmap_Colocalization[0,:,:], dtype=np.bool))
    sns.set(font_scale=0.5)    
    for TME in range(len(heatmap_Colocalization)):
        plt.close('all')
        plt.figure()
        heatmap = sns.heatmap(heatmap_Colocalization[TME,:,:], mask=mask, vmin=heatmap_Colocalization.min(), vmax=heatmap_Colocalization.max(),xticklabels=Marker_Names,yticklabels=Marker_Names, cmap='BrBG')
        heamap = heatmap.get_figure()
        heamap.savefig(dataset.bioInsights_dir_cell_types+cell_type+'/Marker_Colocalization/heatmap_Colocalization_'+abrev+str(TME+1)+'.png',dpi=600) 
        df = pd.DataFrame(heatmap_Colocalization[TME,:,:], columns = Marker_Names)
        df.index = Marker_Names
        df.to_excel(dataset.bioInsights_dir_cell_types+cell_type+'/Marker_Colocalization/heatmap_Colocalization_'+abrev+str(TME+1)+'.xlsx')  

       

def neigh_comp(TC,phenoInd):
    InteractivityVect = np.zeros(len(phenoInd))
    for n_Phen, PH in enumerate(phenoInd):
        PH = [p[0] for p in PH]
        if len(PH)>0:
            for t in TC:
                if t[2] in PH:
                    InteractivityVect[n_Phen]+=1     
    return InteractivityVect

def obtain_neighborhood_composition(dataset,CropConfPheno,CropConfTissueComm):
    '''
    '''

    # Find indices of phenotypes
    phenoInd = []
    for c in CropConfPheno:
        # Find all indices
        phenoInd.append([c_n[3] for c_n in c])        

    # Generate Interactivity matrix
    dict_neigh = []
    for n_Neighbor, TC in enumerate(CropConfTissueComm):
        dict_neigh.append({'TC':TC,'phenoInd':phenoInd})
    result = parallel_process(dict_neigh,neigh_comp,use_kwargs=True,front_num=1,desc='BioInsights: Calculate phenotype abundance whithin neighborhoods') 
    InteractivityMatrix = np.stack(result)

    # Input
    Colormap_Pheno=cm.jet_r(range(0,255,int(255/len(CropConfPheno))))[:,:3]
    Colormap_Neigh=cm.jet_r(range(0,255,int(255/len(CropConfTissueComm))))[:,:3]

    # Save interactivity matrix
    sns.set(font_scale=1.5)
    plt.close('all')
    heatmapInteractivityMatrix_Fig = sns.clustermap(InteractivityMatrix,col_cluster=False, row_cluster=False, row_colors=Colormap_Neigh, col_colors=Colormap_Pheno,xticklabels=['P'+str(i+1) for i in range(len(CropConfPheno))],yticklabels=['N'+str(i+1) for i in range(len(CropConfTissueComm))], linewidths=0.5, cmap="bwr")            
    plt.xlabel("Phenotypes")
    plt.ylabel("Neighborhoods")
    heatmapInteractivityMatrix_Fig.savefig(dataset.bioInsights_dir_cell_types+'Neighborhoods/heatmap_InteractivityMat_Raw.png',dpi=600) 
    plt.close('all')
    InteractivityMatrix[InteractivityMatrix==0]=1e-3
    heatmapInteractivityMatrix_Fig = sns.clustermap(stats.zscore(InteractivityMatrix,axis=1),col_cluster=False, row_cluster=False, row_colors=Colormap_Neigh, col_colors=Colormap_Pheno,xticklabels=['P'+str(i+1) for i in range(len(CropConfPheno))],yticklabels=['N'+str(i+1) for i in range(len(CropConfTissueComm))], linewidths=0.5, cmap="bwr")            
    plt.xlabel("Phenotypes")
    plt.ylabel("Neighborhoods")
    heatmapInteractivityMatrix_Fig.savefig(dataset.bioInsights_dir_cell_types+'Neighborhoods/heatmap_InteractivityMat_Norm.png',dpi=600) 

def Area_to_Neighborhood_to_Phenotype(dataset,clusters,IndexAndClass, ClusteringThreshold):
    
    # Initialize variables
    neigh_to_area = []
    patch_to_pheno = []
    patch_to_neigh = []

    # Obtain clusters per Slide
    for count, idxclster in enumerate(IndexAndClass):           
        try:

            # Load assignment
            neigh_to_area_assignment = np.load(osp.join(dataset.processed_dir_cell_types,'cluster_assignment_Index_{}_ClustLvl_{}.npy'.format(idxclster[1], clusters[-1])))                
            patch_to_neigh_assignment = np.load(osp.join(dataset.processed_dir_cell_types,'cluster_assignmentPerPatch_Index_{}_0_ClustLvl_{}.npy'.format(idxclster[1], clusters[-2])))        
            patch_to_pheno_assignment = np.load(osp.join(dataset.processed_dir_cell_types,'cluster_assignmentPerPatch_Index_{}_0_ClustLvl_{}.npy'.format(idxclster[1], clusters[-3])))                             
            data = dataset.get(idxclster[1],0)

            # Save assignments
            neigh_to_area.append(neigh_to_area_assignment)
            patch_to_neigh.append(patch_to_neigh_assignment[:data.num_nodes,:])            
            patch_to_pheno.append(patch_to_pheno_assignment[:data.num_nodes,:])                        
        except:
            continue

    # Calculate Neighborhoods to areas
    neigh_to_area = np.stack(neigh_to_area) # pat x neigh x areas
    PercTHrsl_area = np.percentile(neigh_to_area,axis=(0,1),q=ClusteringThreshold)
    Area_to_Neigh = np.zeros((clusters[-1],clusters[-2])) 
    for area in range(clusters[-1]):
        Area_to_Neigh[area,:] = np.sum(neigh_to_area[:,:,area]*np.array(neigh_to_area[:,:,area]>PercTHrsl_area[area]),0)

    # Save Neighborhoods to areas heatmap
    plt.close('all')
    plt.figure()
    sns.set(font_scale=1.1)
    row_colors_Area=cm.jet_r(range(0,255,int(255/clusters[2])))[:,:3]
    yticklabels_Area = ['A'+str(i+1) for i in range(clusters[2])]
    col_colors_Neigh=cm.jet_r(range(0,255,int(255/clusters[1])))[:,:3]
    xticklabels_Neigh = ['N'+str(i+1) for i in range(clusters[1])]
    Area_to_Neigh = Area_to_Neigh/Area_to_Neigh.sum(1,keepdims=True)*100    
    Area_to_Neigh[np.isnan(Area_to_Neigh)] = 0
    h_E_Fig = sns.clustermap(Area_to_Neigh.transpose(),col_cluster=True, row_cluster=True, col_colors=row_colors_Area,
                                row_colors=col_colors_Neigh,yticklabels=xticklabels_Neigh,xticklabels=yticklabels_Area, linewidths=0.5, cmap="bwr")            
    h_E_Fig.savefig(dataset.bioInsights_dir_cell_types_Areas + 'heatmap_Neighborhood_composition_of_Areas_Perc_'+str(ClusteringThreshold)+'.png',dpi=600) 
    plt.close('all')
    plt.figure()
    h_E_Fig = sns.clustermap(Area_to_Neigh.transpose(),col_cluster=True, row_cluster=True,vmin=-1, vmax=1, col_colors=row_colors_Area, z_score=1,
                                row_colors=col_colors_Neigh,yticklabels=xticklabels_Neigh,xticklabels=yticklabels_Area, linewidths=0.5, cmap="bwr")            
    h_E_Fig.savefig(dataset.bioInsights_dir_cell_types_Areas + 'heatmap_Neighborhood_composition_of_Areas_Zscore_'+str(ClusteringThreshold)+'.png',dpi=600) 
    df = pd.DataFrame(Area_to_Neigh.transpose(), columns = yticklabels_Area)
    df.index = xticklabels_Neigh
    df.to_excel(dataset.bioInsights_dir_cell_types_Areas + 'heatmap_Neighborhood_composition_of_Areas_Zscore_'+str(ClusteringThreshold)+'.xlsx')  

    # Phenotypes to neighborhoods
    patch_to_neigh = np.concatenate(patch_to_neigh,0) # patches x neigh
    patch_to_pheno = np.concatenate(patch_to_pheno,0) # patches x pheno    
    PercTHrsl_neigh = np.percentile(patch_to_neigh_assignment,axis=0,q=ClusteringThreshold)
    PercTHrsl_pheno = np.percentile(patch_to_pheno,axis=0,q=ClusteringThreshold)        
    for i in range(patch_to_neigh.shape[1]):
        patch_to_neigh[:,i][PercTHrsl_neigh[i]>=patch_to_neigh[:,i]] = 0 
    for i in range(patch_to_pheno.shape[1]):
        patch_to_pheno[:,i][PercTHrsl_pheno[i]>=patch_to_pheno[:,i]] = 0        
    Neigh_to_Pheno = np.matmul(np.transpose(patch_to_neigh),patch_to_pheno)

    # Save heatmap Neigh_to_Pheno
    plt.close('all')
    plt.figure()
    sns.set(font_scale=1.1)
    row_colors_Neigh=cm.jet_r(range(0,255,int(255/clusters[1])))[:,:3]
    yticklabels_Neigh = ['N'+str(i+1) for i in range(clusters[1])]
    col_colors_Pheno=cm.jet_r(range(0,255,int(255/clusters[0])))[:,:3]
    xticklabels_Pheno = ['P'+str(i+1) for i in range(clusters[0])]
    Neigh_to_Pheno = Neigh_to_Pheno/Neigh_to_Pheno.sum(1,keepdims=True)*100
    Neigh_to_Pheno[np.isnan(Neigh_to_Pheno)] = 0
    h_E_Fig = sns.clustermap(Neigh_to_Pheno.transpose(),col_cluster=True, row_cluster=True, col_colors=row_colors_Neigh,
                                row_colors=col_colors_Pheno,yticklabels=xticklabels_Pheno,xticklabels=yticklabels_Neigh, linewidths=0.5, cmap="bwr")            
    h_E_Fig.savefig(dataset.bioInsights_dir_cell_types + 'Neighborhoods/heatmap_Phenotype_composition_of_neighborhoods_Perc_Thrs{}.png'.format(str(ClusteringThreshold)),dpi=600) 
    plt.close('all')
    plt.figure()
    Neigh_to_Pheno[Neigh_to_Pheno==0] = 1e-16
    h_E_Fig = sns.clustermap(Neigh_to_Pheno.transpose(),col_cluster=True, row_cluster=True,vmin=-1, vmax=1, col_colors=row_colors_Neigh, z_score=1,
                                row_colors=col_colors_Pheno,yticklabels=xticklabels_Pheno,xticklabels=yticklabels_Neigh, linewidths=0.5, cmap="bwr")            
    h_E_Fig.savefig(dataset.bioInsights_dir_cell_types + 'Neighborhoods/heatmap_Phenotype_composition_of_neighborhoods_Zscore_Thrs{}.png'.format(str(ClusteringThreshold)),dpi=600) 
    df = pd.DataFrame(Neigh_to_Pheno.transpose(), columns = yticklabels_Neigh)
    df.index = xticklabels_Pheno
    df.to_excel(dataset.bioInsights_dir_cell_types + 'Neighborhoods/heatmap_Phenotype_composition_of_neighborhoods_Perc_Thrs{}.xlsx'.format(str(ClusteringThreshold)))  


    



# -*- coding: utf-8 -*-
"""
plot a pretty confusion matrix with seaborn
Created on Mon Jun 25 14:17:37 2018
@author: Wagner Cipriano - wagnerbhbr - gmail - CEFETMG / MMC
REFerences:
  https://www.mathworks.com/help/nnet/ref/plotconfusion.html
  https://stackoverflow.com/questions/28200786/how-to-plot-scikit-learn-classification-report
  https://stackoverflow.com/questions/5821125/how-to-plot-confusion-matrix-with-string-axis-rather-than-integer-in-python
  https://www.programcreek.com/python/example/96197/seaborn.heatmap
  https://stackoverflow.com/questions/19233771/sklearn-plot-confusion-matrix-with-labels/31720054
  http://scikit-learn.org/stable/auto_examples/model_selection/plot_confusion_matrix.html#sphx-glr-auto-examples-model-selection-plot-confusion-matrix-py
"""

#imports
from pandas import DataFrame
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.font_manager as fm
from matplotlib.collections import QuadMesh
import seaborn as sn


def get_new_fig(fn, figsize=[9,9]):
    """ Init graphics """
    fig1 = plt.figure(fn, figsize)
    ax1 = fig1.gca()   #Get Current Axis
    ax1.cla() # clear existing plot
    return fig1, ax1
#

def configcell_text_and_colors(array_df, lin, col, oText, facecolors, posi, fz, fmt, show_null_values=0):
    """
      config cell text and colors
      and return text elements to add and to dell
      @TODO: use fmt
    """
    text_add = []; text_del = [];
    cell_val = array_df[lin][col]
    tot_all = array_df[-1][-1]
    per = (float(cell_val) / tot_all) * 100
    curr_column = array_df[:,col]
    ccl = len(curr_column)

    #last line  and/or last column
    if(col == (ccl - 1)) or (lin == (ccl - 1)):
        #tots and percents
        if(cell_val != 0):
            if(col == ccl - 1) and (lin == ccl - 1):
                tot_rig = 0
                for i in range(array_df.shape[0] - 1):
                    tot_rig += array_df[i][i]
                per_ok = (float(tot_rig) / cell_val) * 100
            elif(col == ccl - 1):
                tot_rig = array_df[lin][lin]
                per_ok = (float(tot_rig) / cell_val) * 100
            elif(lin == ccl - 1):
                tot_rig = array_df[col][col]
                per_ok = (float(tot_rig) / cell_val) * 100
            per_err = 100 - per_ok
        else:
            per_ok = per_err = 0

        per_ok_s = ['%.2f%%'%(per_ok), '100%'] [per_ok == 100]

        #text to DEL
        text_del.append(oText)

        #text to ADD
        font_prop = fm.FontProperties(weight='bold', size=fz)
        text_kwargs = dict(color='w', ha="center", va="center", gid='sum', fontproperties=font_prop)
        lis_txt = ['%d'%(cell_val), per_ok_s, '%.2f%%'%(per_err)]
        lis_kwa = [text_kwargs]
        dic = text_kwargs.copy(); dic['color'] = 'g'; lis_kwa.append(dic);
        dic = text_kwargs.copy(); dic['color'] = 'r'; lis_kwa.append(dic);
        lis_pos = [(oText._x, oText._y-0.3), (oText._x, oText._y), (oText._x, oText._y+0.3)]
        for i in range(len(lis_txt)):
            newText = dict(x=lis_pos[i][0], y=lis_pos[i][1], text=lis_txt[i], kw=lis_kwa[i])
            #print 'lin: %s, col: %s, newText: %s' %(lin, col, newText)
            text_add.append(newText)
        #print '\n'

        #set background color for sum cells (last line and last column)
        carr = [0.27, 0.30, 0.27, 1.0]
        if(col == ccl - 1) and (lin == ccl - 1):
            carr = [0.17, 0.20, 0.17, 1.0]
        facecolors[posi] = carr

    else:
        if(per > 0):
            txt = '%s\n%.2f%%' %(cell_val, per)
        else:
            if(show_null_values == 0):
                txt = ''
            elif(show_null_values == 1):
                txt = '0'
            else:
                txt = '0\n0.0%'
        oText.set_text(txt)

        #main diagonal
        if(col == lin):
            #set color of the textin the diagonal to white
            oText.set_color('w')
            # set background color in the diagonal to blue
            facecolors[posi] = [0.35, 0.8, 0.55, 1.0]
        else:
            oText.set_color('r')

    return text_add, text_del
#

def insert_totals(df_cm):
    """ insert total column and line (the last ones) """
    sum_col = []
    for c in df_cm.columns:
        sum_col.append( df_cm[c].sum() )
    sum_lin = []
    for item_line in df_cm.iterrows():
        sum_lin.append( item_line[1].sum() )
    df_cm['sum_lin'] = sum_lin
    sum_col.append(np.sum(sum_lin))
    df_cm.loc['sum_col'] = sum_col
    #print ('\ndf_cm:\n', df_cm, '\n\b\n')
#

def pretty_plot_confusion_matrix(df_cm, annot=True, cmap="Oranges", fmt='.2f', fz=11,
      lw=0.5, cbar=False, figsize=[8,8], show_null_values=0, pred_val_axis='y'):
    """
      print conf matrix with default layout (like matlab)
      params:
        df_cm          dataframe (pandas) without totals
        annot          print text in each cell
        cmap           Oranges,Oranges_r,YlGnBu,Blues,RdBu, ... see:
        fz             fontsize
        lw             linewidth
        pred_val_axis  where to show the prediction values (x or y axis)
                        'col' or 'x': show predicted values in columns (x axis) instead lines
                        'lin' or 'y': show predicted values in lines   (y axis)
    """
    if(pred_val_axis in ('col', 'x')):
        xlbl = 'Predicted'
        ylbl = 'Actual'
    else:
        xlbl = 'Actual'
        ylbl = 'Predicted'
        df_cm = df_cm.T

    # create "Total" column
    insert_totals(df_cm)

    #this is for print allways in the same window
    fig, ax1 = get_new_fig('Conf matrix default', figsize)

    #thanks for seaborn
    ax = sn.heatmap(df_cm, annot=annot, annot_kws={"size": fz}, linewidths=lw, ax=ax1,
                    cbar=cbar, cmap=cmap, linecolor='w', fmt=fmt)

    #set ticklabels rotation
    ax.set_xticklabels(ax.get_xticklabels(), rotation = 45, fontsize = 16)
    ax.set_yticklabels(ax.get_yticklabels(), rotation = 25, fontsize = 16)

    # Turn off all the ticks
    for t in ax.xaxis.get_major_ticks():
        t.tick1line.set_visible = False
        t.tick2line.set_visible = False        
    for t in ax.yaxis.get_major_ticks():
        t.tick1line.set_visible = False
        t.tick2line.set_visible = False

    #face colors list
    quadmesh = ax.findobj(QuadMesh)[0]
    facecolors = quadmesh.get_facecolors()

    #iter in text elements
    array_df = np.array( df_cm.to_records(index=False).tolist() )
    text_add = []; text_del = [];
    posi = -1 #from left to right, bottom to top.
    for t in ax.collections[0].axes.texts: #ax.texts:
        pos = np.array( t.get_position()) - [0.5,0.5]
        lin = int(pos[1]); col = int(pos[0]);
        posi += 1
        #print ('>>> pos: %s, posi: %s, val: %s, txt: %s' %(pos, posi, array_df[lin][col], t.get_text()))

        #set text
        txt_res = configcell_text_and_colors(array_df, lin, col, t, facecolors, posi, fz, fmt, show_null_values)

        text_add.extend(txt_res[0])
        text_del.extend(txt_res[1])

    #remove the old ones
    for item in text_del:
        item.remove()
    #append the new ones
    for item in text_add:
        ax.text(item['x'], item['y'], item['text'], **item['kw'])

    #titles and legends
    ax.set_title('Confusion matrix')
    ax.set_xlabel(xlbl, fontsize = 16)
    ax.set_ylabel(ylbl, fontsize = 16)
    plt.tight_layout()  #set layout slim    
    return plt
#

def plot_confusion_matrix_from_data(y_test, predictions, columns=None, annot=True, cmap="Oranges",
      fmt='.2f', fz=18, lw=0.5, cbar=False, figsize=[8,8], show_null_values=0, pred_val_axis='lin'):
    """
        plot confusion matrix function with y_test (actual values) and predictions (predic),
        whitout a confusion matrix yet
    """
    from sklearn.metrics import confusion_matrix
    from pandas import DataFrame

    #data
    if(not columns):
        #labels axis integer:
        ##columns = range(1, len(np.unique(y_test))+1)
        #labels axis string:
        from string import ascii_uppercase
        columns = ['class %s' %(i) for i in list(ascii_uppercase)[0:len(np.unique(y_test))]]

    confm = confusion_matrix(y_test, predictions)
    cmap = 'Oranges'
    fz = 18
    figsize=[9,9]
    show_null_values = 2
    df_cm = DataFrame(confm, index=columns, columns=columns)
    return pretty_plot_confusion_matrix(df_cm, fz=fz, cmap=cmap, figsize=figsize, show_null_values=show_null_values, pred_val_axis=pred_val_axis)

#
#TEST functions
#
def _test_cm():
    #test function with confusion matrix done
    array = np.array( [[13,  0,  1,  0,  2,  0],
                       [ 0, 50,  2,  0, 10,  0],
                       [ 0, 13, 16,  0,  0,  3],
                       [ 0,  0,  0, 13,  1,  0],
                       [ 0, 40,  0,  1, 15,  0],
                       [ 0,  0,  0,  0,  0, 20]])
    #get pandas dataframe
    df_cm = DataFrame(array, index=range(1,7), columns=range(1,7))
    #colormap: see this and choose your more dear
    cmap = 'PuRd'
    pretty_plot_confusion_matrix(df_cm, cmap=cmap)
#

def _test_data_class():
    """ test function with y_test (actual values) and predictions (predic) """
    #data
    y_test = np.array([1,2,3,4,5, 1,2,3,4,5, 1,2,3,4,5, 1,2,3,4,5, 1,2,3,4,5, 1,2,3,4,5, 1,2,3,4,5, 1,2,3,4,5, 1,2,3,4,5, 1,2,3,4,5, 1,2,3,4,5, 1,2,3,4,5, 1,2,3,4,5, 1,2,3,4,5, 1,2,3,4,5, 1,2,3,4,5, 1,2,3,4,5, 1,2,3,4,5, 1,2,3,4,5, 1,2,3,4,5, 1,2,3,4,5, 1,2,3,4,5])
    predic = np.array([1,2,4,3,5, 1,2,4,3,5, 1,2,3,4,4, 1,4,3,4,5, 1,2,4,4,5, 1,2,4,4,5, 1,2,4,4,5, 1,2,4,4,5, 1,2,3,3,5, 1,2,3,3,5, 1,2,3,4,4, 1,2,3,4,1, 1,2,3,4,1, 1,2,3,4,1, 1,2,4,4,5, 1,2,4,4,5, 1,2,4,4,5, 1,2,4,4,5, 1,2,3,4,5, 1,2,3,4,5, 1,2,3,4,5, 1,2,3,4,5])
    """
      Examples to validate output (confusion matrix plot)
        actual: 5 and prediction 1   >>  3
        actual: 2 and prediction 4   >>  1
        actual: 3 and prediction 4   >>  10
    """
    columns = []
    annot = True;
    cmap = 'Oranges';
    fmt = '.2f'
    lw = 0.5
    cbar = False
    show_null_values = 2
    pred_val_axis = 'y'
    #size::
    fz = 12;
    figsize = [9,9];
    if(len(y_test) > 10):
        fz=9; figsize=[14,14];
    return plot_confusion_matrix_from_data(y_test, predic, columns,
      annot, cmap, fmt, fz, lw, cbar, figsize, show_null_values, pred_val_axis)
#


#
#MAIN function
#
if(__name__ == '__main__'):
    print('__main__')
    print('_test_cm: test function with confusion matrix done\nand pause')
    _test_cm()
    plt.pause(5)
    print('_test_data_class: test function with y_test (actual values) and predictions (predic)')
    _test_data_class()


import numpy as np

np.random.seed(0)


class ContrastiveLearningViewGenerator(object):
    """Take two random crops of one image as the query and key."""

    def __init__(self, base_transform, n_views=2):
        self.base_transform = base_transform
        self.n_views = n_views

    def __call__(self, x):
        return [self.base_transform(x) for i in range(self.n_views)]



import torch
import torch.nn.functional as F
from torch_geometric.nn.inits import uniform
from torch_geometric.nn import JumpingKnowledge
import models.pooling as pooling
import models.GNN as GNN
import utilz

class DiffPool_Dense_ATT_Pheno(torch.nn.Module):
    def __init__(self, num_features, num_classes, hidden, num_nodes, clusts, args):
        super(DiffPool_Dense_ATT_Pheno, self).__init__()
        # Initialization
        self.features = num_features
        self.args = args
        num_nodes = clusts[0]
        self.hidden = hidden

        # # LSTM on features
        # if args['LSTMonFeatures']:
        #     n_layers = 6
        #     lstm_layer = nn.LSTM(num_features, floor(hidden/2), 6, batch_first=True)        
        #     hidden_state = torch.randn(n_layers, batch_size, hidden_dim)
        #     cell_state = torch.randn(n_layers, batch_size, hidden_dim)

        # Attention and Readout
        if args['isAttentionLayer']:
            self.attention1 = torch.nn.Linear(hidden,hidden,bias=True)
            self.BatchNormATT = torch.nn.BatchNorm1d(hidden,track_running_stats=False)
            self.attention2 = torch.nn.Linear(hidden,1,bias=True)
        else:
            self.attention1, self.BatchNormATT, self.attention2 = [], [], []
        if args['ReadoutFunction']=='DeepSets' and not args['ObjectiveCluster']:
            self.deepsets1=torch.nn.Linear(hidden,hidden+hidden,bias=True)
            self.deepsets2=torch.nn.Linear(num_nodes,hidden,bias=True)
        else:
            self.deepsets1, self.deepsets2 = [], []

        # Graph Neural Network Embedding
        if args['modeltype']=='SAGE':
            self.embed_pool_block_emb = GNN.GNN(num_features, hidden, hidden, args, mode='SparseMultiplication') if not args['ObjectiveCluster'] else []
            self.embed_pool_block_clust = GNN.GNN(num_features, hidden,  num_nodes, args, mode='SparseMultiplication')
        elif args['modeltype']=='SGC':
            self.embed_pool_block_emb = GNN.Dense_SGC(num_features, hidden, hidden, args) if not args['ObjectiveCluster'] else []
            self.embed_pool_block_clust = GNN.Dense_SGC(num_features, hidden,  num_nodes, args)        
        self.BatchNorm = torch.nn.BatchNorm1d(hidden,track_running_stats=False)
        self.embed_blocks_emb = torch.nn.ModuleList()
        self.embed_blocks_clust = torch.nn.ModuleList()
        self.attention1_blocks = torch.nn.ModuleList()        
        self.attention2_blocks = torch.nn.ModuleList()
        self.BatchNormATT2 = torch.nn.ModuleList()         
        self.deepsets1_blocks=torch.nn.ModuleList()
        self.deepsets2_blocks=torch.nn.ModuleList()

        num_total_clusts = num_nodes
        for i in range(1,len(clusts)):
            numtotalNodes = num_nodes
            num_nodes = clusts[i]
            num_total_clusts += num_nodes
            self.embed_blocks_emb.append(GNN.GNN(hidden, hidden, hidden, args, mode='Multiplication'))
            self.embed_blocks_clust.append(GNN.GNN(hidden, hidden, num_nodes, args, mode='Multiplication'))
            self.attention1_blocks.append(torch.nn.Linear(hidden,hidden,bias=True)) if args['isAttentionLayer'] else self.attention1_blocks.append(torch.nn.Linear(1,1,bias=True))
            self.attention2_blocks.append(torch.nn.Linear(hidden,1,bias=True)) if args['isAttentionLayer'] else self.attention2_blocks.append(torch.nn.Linear(1,1,bias=True))
            self.BatchNormATT2.append(torch.nn.BatchNorm1d(hidden,track_running_stats=False)) if args['isAttentionLayer'] else self.BatchNormATT2.append(torch.nn.Linear(1,1,bias=True))
            if args['ReadoutFunction']=='DeepSets':
                self.deepsets1_blocks.append(torch.nn.Linear(hidden,hidden+hidden,bias=True))
                self.deepsets2_blocks.append(torch.nn.Linear(num_nodes,hidden,bias=True))    
            else:
                self.deepsets1_blocks.append(torch.nn.Linear(1,1,bias=True))
                self.deepsets2_blocks.append(torch.nn.Linear(1,1,bias=True))     

        # Ending layers
        if args['NearestNeighborClassification']:
            self.lin1NNC = torch.nn.Linear(sum([c+(c*(c-1)) for c in clusts]), 10, bias=True)
            self.BNNNC = torch.nn.BatchNorm1d(hidden+hidden,track_running_stats=False)
            self.lin2NNC = torch.nn.Linear(hidden+hidden, 10, bias=True)
        self.jump = JumpingKnowledge(mode='cat')
        self.lin1 = torch.nn.Linear((hidden)*(len(clusts)), hidden+hidden, bias=True) if not args['ObjectiveCluster'] else torch.nn.Linear(sum([c for c in clusts]), hidden+hidden, bias=True) # torch.nn.Linear(sum([c+(c*(c-1)) for c in clusts]), hidden+hidden, bias=True)
        self.BNLast = torch.nn.BatchNorm1d(hidden+hidden,track_running_stats=False)
        self.lin2 = torch.nn.Linear(hidden+hidden, num_classes, bias=True)

        # # MultiTask Learning
        # if args['pearsonCoeffSUP']:
        #     self.pearsonSupMTLearning=torch.zeros((1)).cuda()#torch.nn.Parameter(torch.zeros((1)))
        # if args['pearsonCoeffUNSUP']:
        #     self.pearsonUnsupMTLearning=torch.zeros((1)).cuda()#torch.nn.Parameter(torch.zeros((1)))
        # if args['orthoColor']:
        #     self.orthoColorMTLearning=torch.zeros((1)).cuda()#torch.nn.Parameter(torch.zeros((1)))
        # if args['ortho']:
        #     self.orthoMTLearning=torch.zeros((1)).cuda()#torch.nn.Parameter(torch.zeros((1)))
        # if args['MinCut']:
        #     self.MinCutMTLearning=torch.zeros((1)).cuda()#torch.nn.Parameter(torch.zeros((1)))
        # if args['NearestNeighborClassification']:
        #     self.NNClsfctionMTLearning=torch.zeros((1)).cuda()#torch.nn.Parameter(torch.zeros((1)))
        # self.NNClsfctionMTLearning=torch.zeros((1)).cuda()#torch.nn.Parameter(torch.zeros((1)))
        

    def reset_parameters(self):        
        try:
            self.embed_pool_block_emb.reset_parameters()      
        except:
            self.embed_pool_block_emb=[]
        self.embed_pool_block_clust.reset_parameters()              
        try:
            self.attention1.reset_parameters() 
            self.attention2.reset_parameters()        
        except:
            self.attention1, self.attention2 = [], []
        try: 
            self.deepsets1.reset_parameters()
            self.deepsets2.reset_parameters()
            for deepsets1, deepsets2 in zip(self.deepsets1_blocks, self.deepsets2_blocks):
                deepsets1.reset_parameters()
                deepsets2.reset_parameters()
        except:
            self.deepsets1, self.deepsets2 = [], []
        for block_emb, block_clust, attention1, attention2 in zip(self.embed_blocks_emb,self.embed_blocks_clust,self.attention1_blocks, self.attention2_blocks):
            block_emb.reset_parameters()
            block_clust.reset_parameters()
            attention1.reset_parameters()
            attention2.reset_parameters()
        self.jump.reset_parameters()
        self.lin1.reset_parameters()
        self.lin2.reset_parameters()
        try:
            self.lin1NNC.reset_parameters()
            self.lin2NNC.reset_parameters()
        except:
            self.lin1NNC, self.lin2NNC= [], []        
            

    def attentionLayer(self,x,edge_index,attention1,BatchNormATT,attention2,deepsets1,deepsets2,args,device,attentionVectList,s_interaction):
        if args['isAttentionLayer']:                
            attentionVect = BatchNormATT(F.dropout(F.relu(attention1(x)),p=args['dropoutRate'], training=self.training).transpose(-1,-2)).transpose(-1,-2)                
            attentionVect = BatchNormATT(attentionVect.transpose(-1,-2)).transpose(-1,-2) if attentionVect.shape[0]>1 else attentionVect
            attentionVect = attention2(attentionVect).squeeze(-1)
            attentionVect = F.softmax(attentionVect,dim=-1)
            attentionVect = [torch.where(attentionVect>args['attntnThreshold']/attentionVect.shape[1], attentionVect, torch.tensor([0],dtype=torch.float32).to(device))]                
            x = torch.transpose(torch.transpose(torch.transpose(torch.transpose(x,1,0),0,2)*(attentionVect[-1]),0,2),1,0)
            s_interaction.append(edge_index)
            edge_index = torch.transpose(torch.transpose(torch.transpose(edge_index,1,0)*(attentionVect[-1]),2,0)*(attentionVect[-1]),1,0)
            xs=x.sum(dim=1) if args['ReadoutFunction']=='SUM' else x.max(dim=1)[0]/torch.nonzero(attentionVect[-1]).size(1) if args['ReadoutFunction']=='MAX' else x.sum(dim=1)/torch.nonzero(attentionVect[-1]).size(1)
            attentionVectList.append(attentionVect)
        else:
            s_interaction.append(edge_index)
            xs=x.sum(dim=1) if args['ReadoutFunction']=='SUM' else x.max(dim=1)[0] if args['ReadoutFunction']=='MAX' else deepsets2(deepsets1(x).sum(dim=-1))
            attentionVectList=[]
        return s_interaction, edge_index, x, xs, attentionVectList

    def forward(self, data, device, saveClusters, trainClustering, doClustering,index,Indices,labels,args):

        # First model instance to obtain the interaction-graph
        if doClustering:
            # GNN
            self.s = self.embed_pool_block_clust(data.x, data.edge_index, device, data.num_nodes,args)    
            data.x = [] if args['ObjectiveCluster'] else self.embed_pool_block_emb(data.x, data.edge_index, device, data.num_nodes,args)    
            
            # Apply softmax to Cluster Assignment
            p=0.0
            self.s=F.softmax(self.s,dim=-1)
            self.s = torch.where(self.s>args['attntnThreshold']/self.s.shape[2], self.s, torch.tensor([0],dtype=torch.float32).to(device))
            self.s = [self.s/self.s.sum(dim=-1,keepdim=True)]
            # print(sum(torch.bincount(s[-1][0,:data.num_nodes[0],:].argmax(-1))>0))
            
            # Node-Pooling
            data.x, data.edge_index, ortho_color, pearsonCoeffSUP, pearsonCoeffUnsup, minCut, ortho = pooling.Dense_Pooling(data.y,data.x, data.edge_index, self.s[-1],self.BatchNorm, device, data.num_nodes,doClustering, args,self)  # Regions                        
            # Save Patient-Matrix for later on.            
            if args['ObjectiveCluster'] or args['NearestNeighborClassification']:
                self.sPatient = [torch.zeros(self.s[0].shape[0],self.s[0].shape[-1],dtype=torch.float32).to(device)]
                for i in range(self.s[0].shape[0]):
                    self.sPatient[0][i,:] = (self.s[0][i,:data.num_nodes[i],:].sum(0)/data.num_nodes[i])            
            return data,ortho_color, minCut, ortho, pearsonCoeffSUP, pearsonCoeffUnsup, self.s[-1]            
        else:
            # Use attention in our neural netwwork or not, and apply readout Function to X
            s_interaction, data.edge_index, data.x, ReadoutX, attentionVect = self.attentionLayer(data.x, data.edge_index,self.attention1,self.BatchNormATT,self.attention2, self.deepsets1, self.deepsets2,args,device,[],[])            
            xs = [ReadoutX]            
            minCut=0
            ortho=0
            ortho_color=0
            pearsonCoeffSUP=0
            pearsonCoeffUNSUP=0
            self.edge_index = [data.edge_index]
        for embed, clust, attention1, attention2, batchnorm, deepsets1, deepsets2 in zip(self.embed_blocks_emb, self.embed_blocks_clust, self.attention1_blocks,self.attention2_blocks, self.BatchNormATT2,self.deepsets1_blocks,self.deepsets2_blocks):                                           
            data.x = embed(data.x, data.edge_index,device, None,args) # Feats  
            sux = clust(data.x, data.edge_index, device, None,args) # Clust  
            s_aux = F.softmax(sux,dim=-1)
            s_aux=torch.where(s_aux>args['attntnThreshold']/s_aux.shape[2],s_aux, torch.tensor([0],dtype=torch.float32).to(device))
            s_aux = s_aux/s_aux.sum(dim=-1,keepdim=True)                      
            data.x, data.edge_index, min_next, ortho_next, ortho_color_next, pearsonCoeffSUP_next, pearsonCoeffUNSUP_next = pooling.Sparse_Pooling(data.y,data.x, data.edge_index, s_aux,device,args,self)                        
            self.s.append(s_aux)
            self.edge_index.append(data.edge_index) if args['ObjectiveCluster'] else []
            # Save Patient-Matrix for later on.
            if args['ObjectiveCluster'] or args['NearestNeighborClassification']:
                self.sPatient.append(torch.zeros(s_aux.shape[0],s_aux.shape[-1],dtype=torch.float32).to(device))
                for i in range(self.sPatient[-1].shape[0]):
                    self.sPatient[-1][i,:] = (s_aux[i,:,:].sum(0)/s_aux[i,:,:].shape[0])
            
            # Attention layer or not            
            if not doClustering:
                # Use attention in our neural network or not, and apply readout Function to X                                                                                                                
                s_interaction, data.edge_index, data.x, ReadoutX, attentionVect = self.attentionLayer(data.x,data.edge_index,attention1,batchnorm,attention2,deepsets1,deepsets2,args,device,attentionVect,s_interaction)                
                xs.append(ReadoutX)
                
            # Add unsupevised losses
            minCut += min_next
            ortho += ortho_next
            ortho_color += ortho_color_next
            pearsonCoeffSUP += pearsonCoeffSUP_next
            pearsonCoeffUNSUP += pearsonCoeffUNSUP_next
        # Normalize Losses
        minCut /= len(self.s)-1 if len(self.s)>1 else 1
        ortho /= len(self.s)-1 if len(self.s)>1 else 1
        ortho_color /=len(self.s)-1 if len(self.s)>1 else 1
        pearsonCoeffUNSUP /= len(self.s)-1 if len(self.s)>1 else 1
        pearsonCoeffSUP /= len(self.s)-1 if len(self.s)>1 else 1
        attention_sparseness = 0
        
        # Final Layer
        if not doClustering:            
            if args['ObjectiveCluster']: # Just use S to classify.
                # x=  F.relu(self.lin1(self.jump(self.sPatient))) 
                # x = self.BNLast(x) if x.shape[0]>1 else x
                # x = F.dropout(x, p=args['dropoutRate'], training=self.training)
                # x = self.lin2(x) 
                x = F.relu(self.lin1(torch.cat(self.sPatient,1)))
                x = self.BNLast(x) if x.shape[0]>1 else x
                x = F.dropout(x, p=args['dropoutRate'], training=self.training)
                x = self.lin2(x) 
            else: # Use X to classify
                x = self.jump(xs)
                x = F.relu(self.lin1(x))
                x = self.BNLast(x) if x.shape[0]>1 else x
                x = F.dropout(x, p=args['dropoutRate'], training=self.training)
                x = self.lin2(x)
            if args['NearestNeighborClassification']:                  
                # self.edge_index = [e[torch.nonzero((torch.eye(e.shape[-1],device=device)-1).abs().expand_as(e),as_tuple=True)].view(-1,(e.shape[-1]-1)*e.shape[-1]) for e in self.edge_index] # Extract edges from Adjacency Matrix.
                # self.NNx = torch.cat((self.jump(self.sPatient),self.jump(self.edge_index)),dim=1)
                # self.NNx = self.lin1NNC(self.NNx)
                # self.NNx = self.BNNNC(self.NNx) if self.NNx.shape[0]>1 else self.NNx
                # self.NNx = self.lin2NNC(self.NNx)                
                NN_loss, _, _, _ = utilz.nearestNeighbor_loss(self.training, args, self.sPatient[-1], data.y.long(), self.NNCPosition,index,Indices,labels,device,self)                
                # self.NNx = self.jump(self.s) 
                # 
            else:
                NN_loss=0                            
        else:
            x=torch.Tensor([0,0,0]).to(device)
        return x, ortho_color, minCut, ortho, self.s[1:], s_interaction, attention_sparseness, attentionVect, pearsonCoeffSUP, pearsonCoeffUNSUP, NN_loss
